/*  $Id$

    This program is a part of Neolith project distribution. The Neolith
    project is based on MudOS v22pre5 LPmud driver. Read doc/Copyright
    before you try to use, modify or distribute this program.

    For more information about Neolith project, please visit:

    http://www.es2.muds.neet/neolith
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif	/* HAVE_CONFIG_H */

#ifdef __CYGWIN__
#include <mingw/stdint.h>
#else
#ifdef	HAVE_STDINT_H
#include <stdint.h>
#else	/* ! HAVE_STDINT_H */
typedef	uint32_t unsigned long;
#endif	/* ! HAVE_STDINT_H */
#endif

#include "_adt/crc32.h"
#include "src/std.h"
#include "src/lpc_incl.h"
#include "src/file_incl.h"
#include "src/file.h"
#include "src/comm.h"
#include "src/compiler.h"
#include "src/backend.h"
#include "src/otable.h"
#include "src/stralloc.h"
#include "src/swap.h"
#include "_efuns/call_out.h"
#include "_efuns/ed.h"
#include "LPC/localtime.h"

#include "sprintf.h"
#include "operator.h"
#include "parse.h"
#include "dumpstat.h"

/* get a value for CLK_TCK for use by times() */
#if (defined(TIMES) && !defined(RUSAGE))
/* this may need #ifdef'd to handle different types of machines */
#include <limits.h>
#endif

int call_origin = 0;

static object_t *ob;

int inherits(program_t *, program_t *);
void add_mapping_pair(mapping_t *, char *, int);
void add_mapping_string(mapping_t *, char *, char *);
void add_mapping_object(mapping_t *, char *, object_t *);
void add_mapping_array(mapping_t *, char *, array_t *);
void add_mapping_shared_string(mapping_t *, char *, char *);
int data_size(object_t * ob);
void reload_object(object_t * obj);

#ifdef F_ADD_ACTION
void
f_add_action(void)
{
    int flag;

    if (st_num_arg == 3) {
	flag = (sp--)->u.number;
    } else flag = 0;

    if (sp->type == T_ARRAY){
	int i, n = sp->u.arr->size;
	svalue_t *sv = sp->u.arr->item;

	for (i = 0; i < n; i++){
	    if (sv[i].type == T_STRING){
		add_action(sp-1, sv[i].u.string, flag & 3);
	    }
	}
	free_array((sp--)->u.arr);
    } else {
	add_action((sp-1), sp->u.string, flag & 3);
	free_string_svalue(sp--);
    }
    pop_stack();
}
#endif

#ifdef F_ALL_INVENTORY
void
f_all_inventory(void)
{
    array_t *vec = all_inventory(sp->u.ob, 0);
    free_object(sp->u.ob, "f_all_inventory");
    sp->type = T_ARRAY;
    sp->u.arr = vec;
}
#endif

#ifdef F_ALLOCATE
void
f_allocate(void)
{
    sp->u.arr = allocate_array(sp->u.number);
    sp->type = T_ARRAY;
}
#endif

#ifdef F_ALLOCATE_BUFFER
void
f_allocate_buffer(void)
{
    buffer_t *buf;

    buf = allocate_buffer(sp->u.number);
    if (buf) {
	pop_stack();
	push_refed_buffer(buf);
    } else {
	assign_svalue(sp, &const0);
    }
}
#endif

#ifdef F_ALLOCATE_MAPPING
void
f_allocate_mapping(void)
{
    sp->type = T_MAPPING;
    sp->u.map = allocate_mapping(sp->u.number);
}
#endif

#ifdef F_BIND
void
f_bind(void)
{
    object_t *ob = sp->u.ob;
    funptr_t *old_fp = (sp-1)->u.fp;
    funptr_t *new_fp;
    svalue_t *res;

    if (ob == old_fp->hdr.owner) {
	/* no change */
	free_object(ob, "bind nop");
	sp--;
	return;
    }

    if (old_fp->hdr.type == (FP_LOCAL | FP_NOT_BINDABLE))
	error(_("Local function is not bindable.\n"));
    if (old_fp->hdr.type & FP_NOT_BINDABLE)
	error(_("Function that references global variables is not bindable.\n"));

    /* the object doing the binding */
    if (current_object->flags & O_DESTRUCTED)
	push_number(0);
    else
	push_object(current_object);

    /* the old owner */
    if (old_fp->hdr.owner->flags & O_DESTRUCTED)
	push_number(0);
    else
	push_object(old_fp->hdr.owner);

    /* the new owner */
    push_object(ob);

    res = apply_master_ob(APPLY_VALID_BIND, 3);
    if (!MASTER_APPROVED(res))
	error(_("Permission of binding denied by master object.\n"));

    new_fp = ALLOCATE(funptr_t, TAG_FUNP, "f_bind");
    *new_fp = *old_fp;
    new_fp->hdr.owner = ob; /* one ref from being on stack */
    if (new_fp->hdr.args)
	new_fp->hdr.args->ref++;
    if ((old_fp->hdr.type & 0x0f) == FP_FUNCTIONAL) {
	new_fp->f.functional.prog->func_ref++;
    }

    free_funp(old_fp);
    sp--;
    sp->u.fp = new_fp;
}
#endif

#ifdef F_CACHE_STATS
static void
print_cache_stats(outbuffer_t* ob)
{
    outbuf_add(ob, "Function cache information\n");
    outbuf_add(ob, "-------------------------------\n");
    outbuf_addv(ob, "%% cache hits:    %10.2f\n",
	     100 * ((double) apply_low_cache_hits / apply_low_call_others));
    outbuf_addv(ob, "call_others:     %10lu\n", apply_low_call_others);
    outbuf_addv(ob, "cache hits:      %10lu\n", apply_low_cache_hits);
    outbuf_addv(ob, "cache size:      %10lu\n", APPLY_CACHE_SIZE);
    outbuf_addv(ob, "slots used:      %10lu\n", apply_low_slots_used);
    outbuf_addv(ob, "%% slots used:    %10.2f\n",
		100 * ((double) apply_low_slots_used / APPLY_CACHE_SIZE));
    outbuf_addv(ob, "collisions:      %10lu\n", apply_low_collisions);
    outbuf_addv(ob, "%% collisions:    %10.2f\n",
	     100 * ((double) apply_low_collisions / apply_low_call_others));
}

void f_cache_stats(void)
{
    outbuffer_t ob;

    outbuf_zero(&ob);
    print_cache_stats(&ob);
    outbuf_push(&ob);
}
#endif

#ifdef F_CALL_OTHER
 /* enhanced call_other written 930314 by Luke Mewburn <zak@rmit.edu.au> */
void
f_call_other(void)
{
    svalue_t *arg;
    char *funcname;
    int num_arg = st_num_arg;


    if (current_object->flags & O_DESTRUCTED) {	/* No external calls allowed */
	pop_n_elems(num_arg);
	push_undefined();
	return;
    }
    arg = sp - num_arg + 1;
    if (arg[1].type == T_STRING)
	funcname = arg[1].u.string;
    else {			/* must be T_ARRAY then */
        array_t *v = arg[1].u.arr;

        check_for_destr(v);
        if( (v->size < 1) || !(v->item->type == T_STRING))
            error("call_other: 1st elem of array for arg 2 must be a string\n");
        funcname = v->item->u.string;
        num_arg = 2 + merge_arg_lists(num_arg - 2, v, 1);
    }

    if (arg[0].type == T_OBJECT)
	ob = arg[0].u.ob;
    else if (arg[0].type == T_ARRAY) {
	array_t *ret;

	ret = call_all_other(arg[0].u.arr, funcname, num_arg - 2);
        pop_stack();
        free_array(arg->u.arr);
        sp->u.arr = ret;
	return;
    } else {
        object_t *old_ob;
	ob = find_object(arg[0].u.string);
	if (!(old_ob = ob) || !object_visible(ob))
	    error("call_other() couldn't find object\n");
	ob = old_ob;
    }
    /* Send the remaining arguments to the function. */
#ifdef TRACE
    if (TRACEP(TRACE_CALL_OTHER)) {
	do_trace("Call other ", funcname, "\n");
    }
#endif
    call_origin = ORIGIN_CALL_OTHER;
    if (apply_low(funcname, ob, num_arg - 2) == 0) {	/* Function not found */
	pop_2_elems();
	push_undefined();
	return;
    }
    /*
     * The result of the function call is on the stack.  So is the function
     * name and object that was called, though. These have to be removed.
     */
    free_svalue(--sp, "f_call_other:1");
    free_svalue(--sp, "f_call_other:2");
    *sp = *(sp+2);
    return;
}
#endif

#ifdef F_CALL_OUT
void
f_call_out(void)
{
    svalue_t *arg = sp - st_num_arg + 1;
    int num = st_num_arg - 2;
#ifdef CALLOUT_HANDLES
    int ret;

    if (!(current_object->flags & O_DESTRUCTED)) {
	ret = new_call_out(current_object, arg, arg[1].u.number, num, arg + 2);
	/* args have been transfered; don't free them;
	   also don't need to free the int */
	sp -= num + 1;
    } else {
	ret = 0;
	pop_n_elems(num);
	sp--;
    }
    /* the function */
    free_svalue(sp, "call_out");
    put_number(ret);
#else
    if (!(current_object->flags & O_DESTRUCTED)) {
	new_call_out(current_object, arg, arg[1].u.number, num, arg + 2);
	sp -= num + 1;
    } else {
	pop_n_elems(num);
	sp--;
    }
    free_svalue(sp--, "call_out");
#endif
}
#endif

#ifdef F_CALL_OUT_INFO
void
f_call_out_info(void)
{
    push_refed_array(get_all_call_outs());
}
#endif

#if defined(F_CALL_STACK) || defined(F_ORIGIN)
char *
origin_name(int orig)
{
    /* FIXME: this should use ffs() if available (BSD) */
    int i = 0;
    static char *origins[] = {
	"driver",
	"local",
	"call_other",
	"simul",
	"call_out",
	"efun",
	"function pointer",
	"functional"
    };
    while (orig >>= 1) i++;
    return origins[i];
}
#endif

#ifdef F_CALL_STACK
void
f_call_stack(void)
{
    int i, n = csp - &control_stack[0] + 1;
    array_t *ret;

    if (sp->u.number < 0 || sp->u.number > 3)
	error(_("First argument of call_stack() must be 0, 1, 2, or 3.\n"));

    ret = allocate_empty_array(n);

    switch (sp->u.number) {
    case 0:
	ret->item[0].type = T_STRING;
	ret->item[0].subtype = STRING_MALLOC;
	ret->item[0].u.string = add_slash(current_prog->name);
	for (i = 1; i < n; i++) {
	    ret->item[i].type = T_STRING;
	    ret->item[i].subtype = STRING_MALLOC;
	    ret->item[i].u.string = add_slash((csp - i + 1)->prog->name);
	}
	break;
    case 1:
	ret->item[0].type = T_OBJECT;
	ret->item[0].u.ob = current_object;
	add_ref(current_object, "f_call_stack: curr");
	for (i = 1; i < n; i++) {
	    ret->item[i].type = T_OBJECT;
	    ret->item[i].u.ob = (csp - i + 1)->ob;
	    add_ref((csp - i + 1)->ob, "f_call_stack");
	}
	break;
    case 2:
	for (i = 0; i < n; i++) {
	    ret->item[i].type = T_STRING;
	    if (((csp - i)->framekind & FRAME_MASK) == FRAME_FUNCTION) {
		program_t *prog = (i ? (csp-i+1)->prog : current_prog);
		int index = (csp-i)->fr.table_index;
		compiler_function_t *cfp = &prog->function_table[index];

		ret->item[i].subtype = STRING_SHARED;
		ret->item[i].u.string = cfp->name;
		ref_string(cfp->name);
	    } else {
		ret->item[i].subtype = STRING_CONSTANT;
		ret->item[i].u.string = (((csp - i)->framekind & FRAME_MASK) == FRAME_CATCH) ? "CATCH" : "<function>";
	    }
	}
	break;
    case 3:
	ret->item[0].type = T_STRING;
	ret->item[0].subtype = STRING_CONSTANT;
	ret->item[0].u.string = origin_name(caller_type);

	for (i = 1; i < n; i++) {
	    ret->item[i].type = T_STRING;
	    ret->item[i].subtype = STRING_CONSTANT;
	    ret->item[i].u.string = origin_name((csp-i+1)->caller_type);
	}
	break;
    }
    put_array(ret);
}
#endif

#ifdef F_CAPITALIZE
void
f_capitalize(void)
{
    if (islower(sp->u.string[0])) {
	unlink_string_svalue(sp);
	sp->u.string[0] += 'A' - 'a';
    }
}
#endif

#ifdef F_CHILDREN
void
f_children(void)
{
    array_t *vec;

    vec = children(sp->u.string);
    free_string_svalue(sp);
    put_array(vec);
}
#endif

#ifdef F_CLASSP
void
f_classp(void)
{
    if (sp->type == T_CLASS) {
	free_class(sp->u.arr);
	*sp = const1;
    } else {
	free_svalue(sp, "f_classp");
	*sp = const0;
    }
}
#endif

#ifdef F_CLEAR_BIT
void
f_clear_bit(void)
{
    char *str;
    int len, ind, bit;

    if (sp->u.number > MAX_BITS)
        error("clear_bit() bit requested : %d > maximum bits: %d\n", sp->u.number, MAX_BITS);
    bit = (sp--)->u.number;
    if (bit < 0)
	error("Bad argument 2 (negative) to clear_bit().\n");
    ind = bit / 6;
    bit %= 6;
    len = SVALUE_STRLEN(sp);
    if (ind >= len)
	return;         /* return first arg unmodified */
    unlink_string_svalue(sp);
    str = sp->u.string;

    if (str[ind] > 0x3f + ' ' || str[ind] < ' ')
        error("Illegal bit pattern in clear_bit character %d\n", ind);
    str[ind] = ((str[ind] - ' ') & ~(1 << bit)) + ' ';
}
#endif

#ifdef F_CLONEP
void
f_clonep(void)
{
    if ((sp->type == T_OBJECT) && (sp->u.ob->flags & O_CLONE)){
        free_object(sp->u.ob, "f_clonep");
        *sp = const1;
    } else {
        free_svalue(sp, "f_clonep");
        *sp = const0;
    }
}
#endif

#ifdef F_CLONE_OBJECT
void
f_clone_object(void)
{
    svalue_t *arg = sp - st_num_arg + 1;

    ob = clone_object(arg->u.string, st_num_arg - 1);
    free_string_svalue(sp);
    if (ob) {
	put_unrefed_undested_object(ob, "f_clone_object");
    } else *sp = const0;
}
#endif

#ifdef F_COMMAND
void
f_command(void)
{
    int i;

    i = command_for_object(sp->u.string);
    free_string_svalue(sp);
    put_number(i);
}
#endif

#ifdef F_COMMANDS
void
f_commands(void)
{
    push_refed_array(commands(current_object));
}
#endif

#ifdef F_CP
void
f_cp(void)
{
    int i;

    i = copy_file(sp[-1].u.string, sp[0].u.string);
    free_string_svalue(sp--);
    free_string_svalue(sp);
    put_number(i);
}
#endif

#ifdef F_CRC32
void
f_crc32(void)
{
    int len;
    unsigned char *buf;
    uint32_t crc;

    if (sp->type == T_STRING) {
        len = SVALUE_STRLEN(sp);
        buf = (unsigned char *) sp->u.string;
    } else if (sp->type == T_BUFFER) {
        len = sp->u.buf->size;
        buf = sp->u.buf->item;
    } else {
        bad_argument(sp, T_STRING | T_BUFFER, 1, F_CRC32);
        return;
    }
    crc = compute_crc32(buf, len);
    free_svalue(sp, "f_crc32");
    put_number(crc);
}
#endif

#ifdef F_CTIME
void
f_ctime(void)
{
    char *cp, *nl, *p;
    int l;

    cp = ctime( (time_t*) &sp->u.number);

    if ((nl = strchr(cp, '\n')))
	l = nl - cp;
    else
	l = strlen(cp);

    p = new_string(l, "f_ctime");
    strncpy(p, cp, l);
    p[l] = '\0';
    put_malloced_string(p);
}
#endif

#ifdef F_DEEP_INHERIT_LIST
void
f_deep_inherit_list(void)
{
    array_t *vec;

    if (!(sp->u.ob->flags & O_SWAPPED)) {
        vec = deep_inherit_list(sp->u.ob);
    } else {
        vec = &the_null_array;
    }
    free_object(sp->u.ob, "f_deep_inherit_list");
    put_array(vec);
}
#endif

#ifdef DEBUG_MACRO

#ifdef F_SET_DEBUG_LEVEL
void
f_set_debug_level(void)
{
    debug_level = (sp--)->u.number;
}
#endif

#endif

#ifdef F_DEEP_INVENTORY
void
f_deep_inventory(void)
{
    array_t *vec;

    vec = deep_inventory(sp->u.ob, 0);
    free_object(sp->u.ob, "f_deep_inventory");
    put_array(vec);
}
#endif

#ifdef F_DESTRUCT
void
f_destruct(void)
{
    destruct_object(sp->u.ob);
    sp--; /* Ok since the object was removed from the stack */
}
#endif

#ifdef F_DISABLE_COMMANDS
void
f_disable_commands(void)
{
    enable_commands(0);
}
#endif

#ifdef F_DUMPALLOBJ
void
f_dumpallobj(void)
{
    if (st_num_arg) {
        dumpstat(sp->u.string);
        free_string_svalue(sp--);
    } else {
        dumpstat("/OBJ_DUMP");
    }
}
#endif

#ifdef F_ED
void
f_ed(void)
{
    if (!command_giver || !command_giver->interactive) {
	pop_n_elems(st_num_arg);
        return;
    }

    if (!st_num_arg) {
        /* ed() */
        ed_start(0, 0, 0, 0, 0);
    } else if (st_num_arg == 1) {
        /* ed(fname) */
        ed_start(sp->u.string, 0, 0, 0, 0);
	pop_stack();
    } else if (st_num_arg == 2) {
        /* ed(fname,exitfn) */
        ed_start((sp - 1)->u.string, 0, sp->u.string, 0, current_object);
        pop_2_elems();
    } else if (st_num_arg == 3) {
        /* ed(fname,exitfn,restricted) / ed(fname,writefn,exitfn) */
        if (sp->type == T_NUMBER) {
            ed_start((sp - 2)->u.string, 0, (sp - 1)->u.string, sp->u.number,
                     current_object);
	} else if (sp->type == T_STRING) {
            ed_start((sp - 2)->u.string, (sp - 1)->u.string, sp->u.string, 0,
                     current_object);
	} else {
            bad_argument(sp, T_NUMBER | T_STRING, 3, F_ED);
	}
        pop_3_elems();
    } else {                    /* st_num_arg == 4 */
        /* ed(fname,writefn,exitfn,restricted) */
        if (!((sp - 1)->type == T_STRING))
            bad_argument(sp - 1, T_STRING, 3, F_ED);
        if (!(sp->type == T_NUMBER))
            bad_argument(sp, T_NUMBER, 4, F_ED);
        ed_start((sp - 3)->u.string, (sp - 2)->u.string, (sp - 1)->u.string,
                 sp->u.number, current_object);
	pop_n_elems(4);
    }
}
#endif

#ifdef F_ED_CMD
void f_ed_cmd(void)
{
    char *res;

    if (current_object->flags & O_DESTRUCTED)
	error("destructed objects can't use ed.\n");

    if (!(current_object->flags & O_IN_EDIT))
	error("ed_cmd() called with no ed session active.\n");

    res = object_ed_cmd(current_object, sp->u.string);

    free_string_svalue(sp);
    if (res) {
	sp->subtype = STRING_MALLOC;
	sp->u.string = res;
    } else {
	sp->subtype = STRING_CONSTANT;
	sp->u.string = "";
    }
}
#endif

#ifdef F_ED_START
void f_ed_start(void)
{
    char *res;
    char *fname;
    int restr = 0;

    if (st_num_arg == 2)
	restr = (sp--)->u.number;

    if (st_num_arg)
	fname = sp->u.string;
    else
	fname = 0;

    if (current_object->flags & O_DESTRUCTED)
	error("destructed objects can't use ed.\n");

    if (current_object->flags & O_IN_EDIT)
	error("ed_start() called while an ed session is already started.\n");

    res = object_ed_start(current_object, fname, restr);

    if (fname) free_string_svalue(sp);
    else ++sp;

    if (res) {
	sp->subtype = STRING_MALLOC;
	sp->u.string = res;
    } else {
	sp->subtype = STRING_CONSTANT;
	sp->u.string = "";
    }
}
#endif

#ifdef F_ENABLE_COMMANDS
void
f_enable_commands(void)
{
    enable_commands(1);
}
#endif

#ifdef F_ENABLE_WIZARD
void
f_enable_wizard(void)
{
    if (current_object->interactive)
	current_object->flags |= O_IS_WIZARD;
}
#endif

#ifdef F_ERROR
void
f_error(void)
{
    int l = SVALUE_STRLEN(sp);
    char err_buf[2048];

    if (sp->u.string[l - 1] == '\n')
	l--;
    if (l > 2045) l = 2045;

    err_buf[0] = '*';
    strncpy(err_buf + 1, sp->u.string, l);
    err_buf[l + 1] = '\n';
    err_buf[l + 2] = 0;

    error_handler(err_buf);
}
#endif

#ifdef F_DISABLE_WIZARD
void
f_disable_wizard(void)
{
    if (current_object->interactive)
	current_object->flags &= ~O_IS_WIZARD;
}
#endif

#ifdef F_ENVIRONMENT
void
f_environment(void)
{
    if (st_num_arg) {
        if ((ob = sp->u.ob)->flags & O_DESTRUCTED)
            error(_("environment() called on destructed object.\n"));
        ob = ob->super;
        free_object((sp--)->u.ob, "f_environment");
    } else if (!(current_object->flags & O_DESTRUCTED))
        ob = current_object->super;
    else error(_("environment() called on destructed object.\n"));

    if (ob) push_object(ob);
    else push_number(0);
}
#endif

#ifdef F_EXEC
void
f_exec(void)
{
    int i;

    i = replace_interactive((sp - 1)->u.ob, sp->u.ob);

    /* They might have been destructed */
    if (sp->type == T_OBJECT)
	free_object(sp->u.ob, "f_exec:1");
    if ((--sp)->type == T_OBJECT)
	free_object(sp->u.ob, "f_exec:2");
    put_number(i);
}
#endif

#ifdef F_EXPLODE
void
f_explode(void)
{
    array_t *vec;

    vec = explode_string((sp - 1)->u.string, SVALUE_STRLEN(sp-1),
			 sp->u.string, SVALUE_STRLEN(sp));
    free_string_svalue(sp--);
    free_string_svalue(sp);
    put_array(vec);
}
#endif

#ifdef F_FILE_NAME
void
f_file_name(void)
{
    char *res;

    /* This function now returns a leading '/' */
    res = (char *) add_slash(sp->u.ob->name);
    free_object(sp->u.ob, "f_file_name");
    put_malloced_string(res);
}
#endif

#ifdef F_FILE_SIZE
void
f_file_size(void)
{
    int i = file_size(sp->u.string);
    free_string_svalue(sp);
    put_number(i);
}
#endif

#ifdef F_FILTER
void
f_filter(void)
{
    svalue_t *arg = sp - st_num_arg + 1;

    if (arg->type == T_MAPPING) filter_mapping(arg, st_num_arg);
    else filter_array(arg, st_num_arg);
}
#endif

#ifdef F_FIND_CALL_OUT
void
f_find_call_out(void)
{
    int i;
#ifdef CALLOUT_HANDLES
    if (sp->type == T_NUMBER) {
	i = find_call_out_by_handle(sp->u.number);
    } else { /* T_STRING */
#endif
	i = find_call_out(current_object, sp->u.string);
	free_string_svalue(sp);
#ifdef CALLOUT_HANDLES
    }
#endif
    put_number(i);
}
#endif

#ifdef F_FIND_LIVING
void
f_find_living(void)
{
    ob = find_living_object(sp->u.string, 0);
    free_string_svalue(sp);
    /* safe b/c destructed objects have had their living names removed */
    if (ob) { put_unrefed_undested_object(ob, "find_living"); }
    else *sp = const0;
}
#endif

#ifdef F_FIND_OBJECT
void
f_find_object(void)
{
    if ((sp--)->u.number)
        ob = find_object(sp->u.string);
    else
        ob = find_object2(sp->u.string);
    free_string_svalue(sp);
    if (ob) {
	object_t *old_ob = ob;
	/* object_visible might change ob, a global - Sym */
        if (object_visible(ob)) {
	  /* find_object only returns undested objects */
	  put_unrefed_undested_object(old_ob, "find_object");
	}
        else *sp = const0;
    } else *sp = const0;
}
#endif

#ifdef F_FIND_PLAYER
void
f_find_player(void)
{
    ob = find_living_object(sp->u.string, 1);
    free_string_svalue(sp);
    if (ob) {
      put_unrefed_undested_object(ob, "find_player");
    }
    else *sp = const0;
}
#endif

#ifdef F_FUNCTION_PROFILE
/* f_function_profile: John Garnett, 1993/05/31, 0.9.17.3 */
void
f_function_profile(void)
{
    array_t *vec;
    mapping_t *map;
    program_t *prog;
    int nf, j;

    ob = sp->u.ob;
    if (ob->flags & O_SWAPPED) {
	load_ob_from_swap(ob);
    }
    prog = ob->prog;
    nf = prog->num_functions_defined;
    vec = allocate_empty_array(nf);
    for (j = 0; j < nf; j++) {
	map = allocate_mapping(3);
	add_mapping_pair(map, "calls", prog->function_table[j].calls);
	add_mapping_pair(map, "self", prog->function_table[j].self
			 - prog->function_table[j].children);
	add_mapping_pair(map, "children", prog->function_table[j].children);
	add_mapping_shared_string(map, "name", prog->function_table[j].name);
	vec->item[j].type = T_MAPPING;
	vec->item[j].u.map = map;
    }
    free_object(ob, "f_function_profile");
    put_array(vec);
}
#endif

#ifdef F_FUNCTION_EXISTS
void
f_function_exists(void)
{
    char *str, *res;
    int l;
    object_t *ob;
    int flag = 0;

    if (st_num_arg > 1) {
	if (st_num_arg > 2)
	    flag = (sp--)->u.number;
	ob = (sp--)->u.ob;
	free_object(ob, "f_function_exists");
    } else {
	if (current_object->flags & O_DESTRUCTED) {
	    free_string_svalue(sp);
	    *sp = const0;
	    return;
	}
	ob = current_object;
    }

    str = function_exists(sp->u.string, ob, flag);
    free_string_svalue(sp);
    if (str) {
	l = SHARED_STRLEN(str) - 2; /* no .c */
	res = new_string(l + 1, "function_exists");
	res[0] = '/';
	strncpy(res + 1, str, l);
	res[l + 1] = 0;

	sp->subtype = STRING_MALLOC;
        sp->u.string = res;
    } else *sp = const0;
}
#endif

#ifdef F_GENERATE_SOURCE
void f_generate_source(void)
{
    int i;

    if (st_num_arg == 2) {
	i = generate_source(sp - 1, sp->u.string);
	pop_stack();
    } else
	i = generate_source(sp, 0);
    free_svalue(sp, "f_generate_source");
    put_number(i);
}
#endif

#ifdef F_GET_CHAR
void
f_get_char(void)
{
    svalue_t *arg;
    int i, tmp;
    int flag;

    arg = sp - st_num_arg + 1;  /* Points arg at first argument. */
    if (st_num_arg == 1 || !(arg[1].type == T_NUMBER)) {
        tmp = 0;
        flag = 0;
    } else {
        tmp = 1;
        st_num_arg--;           /* Don't count the flag as an arg */
        flag = arg[1].u.number;
    }
    st_num_arg--;
    i = get_char(arg, flag, st_num_arg, &arg[1 + tmp]);
    free_svalue(arg, "f_get_char");
    (sp = arg)->type = T_NUMBER;
    sp->u.number = i;
}
#endif

#ifdef F_GET_CONFIG
void
f_get_config(void)
{
    if (!get_config_item(sp, sp))
	error("Bad argument to get_config()\n");
}
#endif

#ifdef F_GET_DIR
void
f_get_dir(void)
{
    array_t *vec;

    vec = get_dir((sp - 1)->u.string, sp->u.number);
    free_string_svalue(--sp);
    if (vec) { put_array(vec); }
    else *sp = const0;
}
#endif

#ifdef F_IMPLODE
void
f_implode(void)
{
    array_t *arr;
    int flag;
    svalue_t *args;

    if (st_num_arg == 3) {
	args = (sp - 2);
	if (args[1].type == T_STRING)
	    error("Third argument to implode() is illegal with implode(array, string)\n");
	flag = 1;
    } else {
	args = (sp - 1);
	flag = 0;
    }
    arr = args->u.arr;
    check_for_destr(arr);

    if (args[1].type == T_STRING) {
	/* st_num_arg == 2 here */
	char *str;

	str = implode_string(arr, sp->u.string,
			     SVALUE_STRLEN(sp));
	free_string_svalue(sp--);
	free_array(arr);
	put_malloced_string(str);
    } else { /* function */
	funptr_t *funp = args[1].u.fp;

	/* this pulls the extra arg off the stack if it exists */
	implode_array(funp, arr, args, flag);
	sp--;
	free_funp(funp);
	free_array(arr);
    }
}
#endif

#ifdef F_IN_EDIT
void
f_in_edit(void)
{
    char *fn;
    ed_buffer_t *eb = 0;

#ifdef OLD_ED
    if (sp->u.ob->interactive)
	eb = sp->u.ob->interactive->ed_buffer;
#else
    if (sp->u.ob->flags & O_IN_EDIT)
	eb = find_ed_buffer(sp->u.ob);
#endif
    if (eb && (fn = eb->fname)) {
	free_object(sp->u.ob, "f_in_edit:1");
	put_constant_string(fn); /* is this safe?  - Beek */
	return;
    }
    free_object(sp->u.ob, "f_in_edit:1");
    *sp = const0;
    return;
}
#endif

#ifdef F_IN_INPUT
void
f_in_input(void)
{
    int i;

    i = sp->u.ob->interactive && sp->u.ob->interactive->input_to;
    free_object(sp->u.ob, "f_in_input");
    put_number(i != 0);
}
#endif

#ifdef F_INHERITS
int
inherits(program_t* prog, program_t* thep)
{
    int j, k = prog->num_inherited;
    program_t *pg;

    for (j = 0; j < k; j++) {
        if ((pg = prog->inherit[j].prog) == thep)
            return 1;
        if (!strcmp(pg->name, thep->name))
            return 2;
        if (inherits(pg, thep))
            return 1;
    }
    return 0;
}

void
f_inherits(void)
{
    object_t *ob, *base;
    int i;

    base = (sp--)->u.ob;
    ob = find_object2(sp->u.string);
    if (!ob || (ob->flags & O_SWAPPED)) {
	free_object(base, "f_inherits");
        assign_svalue(sp, &const0);
        return;
    }
    if (base->flags & O_SWAPPED)
        load_ob_from_swap(base);
    i = inherits(base->prog, ob->prog);
    free_object(base, "f_inherits");
    free_string_svalue(sp);
    put_number(i);
}
#endif

#ifdef F_SHALLOW_INHERIT_LIST
void
f_shallow_inherit_list(void)
{
    array_t *vec;

    if (!(sp->u.ob->flags & O_SWAPPED)) {
        vec = inherit_list(sp->u.ob);
    } else {
        vec = &the_null_array;
    }
    free_object(sp->u.ob, "f_inherit_list");
    put_array(vec);
}
#endif

#ifdef F_INPUT_TO
void
f_input_to(void)
{
    svalue_t *arg;
    int i, tmp;
    int flag;

    arg = sp - st_num_arg + 1;  /* Points arg at first argument. */
    if ((st_num_arg == 1) || !(arg[1].type == T_NUMBER)) {
        tmp = flag = 0;
    } else {
        tmp = 1;
        st_num_arg--;           /* Don't count the flag as an arg */
        flag = arg[1].u.number;
    }
    st_num_arg--;               /* Don't count the name of the func either. */
    i = input_to(arg, flag, st_num_arg, &arg[1 + tmp]);
    free_svalue(arg, "f_input_to");
    (sp = arg)->type = T_NUMBER;
    sp->u.number = i;
}
#endif

#ifdef F_INTERACTIVE
void
f_interactive(void)
{
    int i;

    i = (sp->u.ob->interactive != 0);
    free_object(sp->u.ob, "f_interactive");
    put_number(i);
}
#endif

#ifdef F_INTP
void
f_intp(void)
{
    if (sp->type == T_NUMBER) sp->u.number = 1;
    else {
	free_svalue(sp, "f_intp");
        put_number(0);
    }
}
#endif

#ifdef F_FUNCTIONP
void
f_functionp(void)
{
    int i;

    if (sp->type == T_FUNCTION) {
        i = sp->u.fp->hdr.type;
        if (sp->u.fp->hdr.args)
	    i |= FP_HAS_ARGUMENTS;
	if (sp->u.fp->hdr.owner->flags & O_DESTRUCTED)
	    i |= FP_OWNER_DESTED;
        free_funp(sp->u.fp);
        put_number(i);
        return;
    }
    assign_svalue(sp, &const0);
}
#endif

#ifdef F_KEYS
void
f_keys(void)
{
    array_t *vec;

    vec = mapping_indices(sp->u.map);
    free_mapping(sp->u.map);
    put_array(vec);
}
#endif

#ifdef F_VALUES
void
f_values(void)
{
    array_t *vec;

    vec = mapping_values(sp->u.map);
    free_mapping(sp->u.map);
    put_array(vec);
}
#endif

#ifdef F_LINK
void
f_link(void)
{
    svalue_t *ret;
    int i;

    push_svalue(sp - 1);
    push_svalue(sp);
    ret = apply_master_ob(APPLY_VALID_LINK, 2);
    if (MASTER_APPROVED(ret))
        i = do_rename((sp - 1)->u.string, sp->u.string, F_LINK);
    else
	i = 0;
    (--sp)->type = T_NUMBER;
    sp->u.number = i;
    sp->subtype = 0;
}
#endif				/* F_LINK */

#ifdef F_LIVING
void
f_living(void)
{
    if (sp->u.ob->flags & O_ENABLE_COMMANDS){
        free_object(sp->u.ob, "f_living:1");
        *sp = const1;
    }
    else {
        free_object(sp->u.ob, "f_living:2");
        *sp = const0;
    }
}
#endif

#ifdef F_LIVINGS
void
f_livings(void)
{
    push_refed_array(livings());
}
#endif

#ifdef F_LOWER_CASE
void
f_lower_case(void)
{
    register char *str;

    str = sp->u.string;
    /* find first upper case letter, if any */
    for (; *str; str++) {
	if (isupper(*str)) {
	    int l = str - sp->u.string;
	    unlink_string_svalue(sp);
	    str = sp->u.string + l;
	    *str += 'a' - 'A';
	    for (str++; *str; str++) {
		if (isupper(*str))
		    *str += 'a' - 'A';
	    }
	    return;
	}
    }
}
#endif

#ifdef F_LPC_INFO
void f_lpc_info(void)
{
    outbuffer_t out;

    interface_t **p = interface;
    object_t *ob;

    outbuf_zero(&out);
    outbuf_addv(&out, "%30s  Loaded  Using compiled program\n", "Program");
    while (*p) {
	outbuf_addv(&out, "%30s: ", (*p)->fname);
	ob = lookup_object_hash((*p)->fname);
	if (ob) {
	    if (ob->flags & O_COMPILED_PROGRAM) {
		outbuf_add(&out, " No\n");
	    } else if (ob->flags & O_SWAPPED) {
		outbuf_add(&out, " Yes      Swapped\n");
	    } else if (ob->prog->program_size == 0) {
		outbuf_add(&out, " Yes      Yes\n");
	    } else {
		outbuf_add(&out, " Yes      No\n");
	    }
	} else {
	    outbuf_add(&out, "Something REALLY wierd happened; no record of the object.\n");
	}
	p++;
    }
    outbuf_push(&out);
}
#endif

#ifdef F_MALLOC_STATUS
void f_malloc_status(void)
{
    outbuffer_t ob;

    outbuf_zero(&ob);

#ifdef BSDMALLOC
    outbuf_add(&ob, "Using BSD malloc");
#endif
#ifdef SMALLOC
    outbuf_add(&ob, "Using Smalloc");
#endif
#ifdef SYSMALLOC
    outbuf_add(&ob, "Using system malloc");
#endif
    outbuf_add(&ob, ".\n");
#ifdef DO_MSTATS
    show_mstats(&ob, "malloc_status()");
#endif
#if (defined(WRAPPEDMALLOC) || defined(DEBUGMALLOC))
    dump_malloc_data(&ob);
#endif
    outbuf_push(&ob);
}
#endif

#ifdef F_MAP_DELETE
void
f_map_delete(void)
{
    mapping_delete((sp - 1)->u.map, sp);
    pop_stack();
#ifndef COMPAT_32
    free_mapping((sp--)->u.map);
#endif
}
#endif

#ifdef F_MAPP
void
f_mapp(void)
{
    if (sp->type == T_MAPPING){
        free_mapping(sp->u.map);
        *sp = const1;
    } else {
        free_svalue(sp, "f_mapp");
        *sp = const0;
    }
}
#endif

#ifdef F_MAP
void
f_map(void)
{
    svalue_t *arg = sp - st_num_arg + 1;

    if (arg->type == T_MAPPING) map_mapping(arg, st_num_arg);
    else if (arg->type == T_ARRAY) map_array(arg, st_num_arg);
    else map_string(arg, st_num_arg);
}
#endif

#ifdef F_MASTER
void
f_master(void)
{
    if (!master_ob)
	push_number(0);
    else
	push_object(master_ob);
}
#endif

/*
This efun searches a mapping for a path.  Each key is assumed to be a
string.  The value is completely arbitrary.  The efun finds the largest
matching path in the mapping.  Keys ended in '/' are assumed to match
paths with character that follow the '/', i.e. / is a wildcard for anything
below this directory.  DO NOT CHANGE THIS EFUN TIL YOU UNDERSTAND IT.  It
catches folks by suprise at first, but it is coded the way it is for a reason.
It effectively implements the search loop in TMI's access object as a single
efun.

Cygnus

* Changed to improve the way ES2 uses by Annihilator (05/13/2000)

*/
#ifdef F_MATCH_PATH
void
f_match_path(void)
{
    svalue_t *value;
    register char *src, *dst;
    svalue_t *nvalue;
    mapping_t *map;
    char *tmpstr;

    value = &const0u;

    tmpstr = DMALLOC(SVALUE_STRLEN(sp) + 1, TAG_STRING, "match_path");

    src = sp->u.string;
    dst = tmpstr;
    map = (sp - 1)->u.map;

    while (*src != '\0') {
	while (*src != '/' && *src != '\0')
	    *dst++ = *src++;
	if (*src == '/') {
	    while (*++src == '/');
	    if( dst == tmpstr ) continue;
	}
	*dst = '\0';
	nvalue = find_string_in_mapping(map, tmpstr);

	value = nvalue;

	if( value == &const0u ) break;
	if( value->type != T_MAPPING ) {
	    if( *src != '\0' ) value = &const0u;
	    break;
	}
	map = value->u.map;
	dst = tmpstr;
    }

    FREE(tmpstr);
    /* Don't free mapping first, in case sometimes one uses a ref 1 mapping */
    /* Randor - 5/29/94 */
    free_string_svalue(sp--);
    map = sp->u.map;
    assign_svalue_no_free(sp, value);
    free_mapping(map);
}
#endif				/* F_MATCH_PATH */

#ifdef F_MEMBER_ARRAY
void
f_member_array(void)
{
    array_t *v;
    int i;

    if (st_num_arg > 2) {
        i = (sp--)->u.number;
        if (i<0) bad_arg(3, F_MEMBER_ARRAY);
    } else i = 0;

    if (sp->type == T_STRING) {
        char *res;
        CHECK_TYPES(sp-1, T_NUMBER, 1, F_MEMBER_ARRAY);
        if (i > SVALUE_STRLEN(sp)) error("Index to start search from in member_array() is > string length.\n");
        if ((res = strchr(sp->u.string + i, (sp-1)->u.number)))
            i = res - sp->u.string;
        else
            i = -1;
        free_string_svalue(sp--);
    } else {
        int size = (v = sp->u.arr)->size;
        svalue_t *sv;
	svalue_t *find;
	int flen = 0;

        find = (sp - 1);
	/* optimize a bit */
	if (find->type == T_STRING) {
	    /* *not* COUNTED_STRLEN() which can do a (costly) strlen() call */
	    if (find->subtype & STRING_COUNTED)
		flen = MSTR_SIZE(find->u.string);
	    else flen = 0;
	}

        for (; i < size; i++) {
            switch (find->type|(sv= v->item + i)->type) {
            case T_STRING:
		if (flen && (sv->subtype & STRING_COUNTED)
		    && flen != MSTR_SIZE(sv->u.string))
		    continue;
                if (strcmp(find->u.string, sv->u.string)) continue;
		break;
            case T_NUMBER:
                if (find->u.number == sv->u.number) break;
                continue;
            case T_REAL:
                if (find->u.real == sv->u.real) break;
                continue;
            case T_ARRAY:
                if (find->u.arr == sv->u.arr) break;
                continue;
            case T_OBJECT:
	    {
		if (sv->u.ob->flags & O_DESTRUCTED){
		    assign_svalue(sv, &const0);
		    continue;
		}
                if (find->u.ob == sv->u.ob) break;
                continue;
	    }
            case T_MAPPING:
                if (find->u.map == sv->u.map) break;
                continue;
            case T_FUNCTION:
                if (find->u.fp == sv->u.fp) break;
                continue;
            case T_BUFFER:
                if (find->u.buf == sv->u.buf) break;
                continue;
            default:
                if (sv->type == T_OBJECT && sv->u.ob->flags & O_DESTRUCTED){
                    assign_svalue(sv, &const0);
                    if (find->type == T_NUMBER && !find->u.number) break;
		}
                continue;
	    }
            break;
	}
        if (i == size)
            i = -1;                     /* Return -1 for failure */
        free_array(v);
	free_svalue(find, "f_member_array");
	sp--;
    }
    put_number(i);
}
#endif

#ifdef F_MESSAGE
void
f_message(void)
{
    array_t *use = NULL, *avoid;
    int num_arg = st_num_arg;
    svalue_t *args;

    args = sp - num_arg + 1;
    switch (args[2].type)	/* ˬd message ǰeHO */
    {
    case T_OBJECT:
    case T_STRING:
	use = allocate_empty_array(1);
	use->item[0] = args[2];
	args[2].type = T_ARRAY;
	args[2].u.arr = use;
	break;
    case T_ARRAY:
	use = args[2].u.arr;
	break;
    default:
	bad_argument(&args[2], T_OBJECT | T_STRING | T_ARRAY, 3, F_MESSAGE);
    }

    if (num_arg == 4) {
	switch (args[3].type) {	/* ˬd message ǰe~HO */
	case T_OBJECT:
	    avoid = allocate_empty_array(1);
	    avoid->item[0] = args[3];
	    args[3].type = T_ARRAY;
	    args[3].u.arr = avoid;
	    break;
	case T_ARRAY:
	    avoid = args[3].u.arr;
	    break;
	default:
	    avoid = &the_null_array;
	}
    } else
	avoid = &the_null_array;

    if( use != NULL ) do_message(&args[0], &args[1], use, avoid, 1);
    pop_n_elems(num_arg);
}
#endif

#ifdef F_MKDIR
void
f_mkdir(void)
{
    char *path;

    path = check_valid_path(sp->u.string, current_object, "mkdir", 1);
    if (!path || OS_mkdir(path, 0770) == -1){
        free_string_svalue(sp);
        *sp = const0;
    }
    else{
        free_string_svalue(sp);
        *sp = const1;
    }
}
#endif

#ifdef F_MOVE_OBJECT
void
f_move_object(void)
{
    object_t *o1, *o2;

    /* get destination */
    if (sp->type == T_OBJECT)
	o2 = sp->u.ob;
    else {
	if (!(o2 = find_object(sp->u.string)) || !object_visible(o2))
	    error("move_object failed: could not find destination\n");
    }

    if ((o1 = current_object)->flags & O_DESTRUCTED)
	error("move_object(): can't move a destructed object\n");

    move_object(o1, o2);
    pop_stack();
}
#endif

#ifdef F_MUD_STATUS
void f_mud_status(void)
{
    int tot, res, verbose = 0;
    outbuffer_t ob;

    outbuf_zero(&ob);
    verbose = (sp--)->u.number;

    if (reserved_area)
	res = RESERVED_SIZE;
    else
	res = 0;

    if (verbose) {
	char dir_buf[PATH_MAX];
	FILE *testfp;

	if ((testfp = fopen(".mudos_test_file", "w"))) {
	    fclose(testfp);
	    outbuf_add(&ob, "Open-file-test succeeded.\n");
	    unlink(".mudos_test_file");
	} else {
	    outbuf_addv(&ob, "Open file test failed: %s\n", strerror(errno));
	}

	outbuf_addv(&ob, "current working directory: %s\n\n",
		    getcwd(dir_buf, PATH_MAX));
	outbuf_add(&ob, "add_message statistics\n");
	outbuf_add(&ob, "------------------------------\n");
	outbuf_addv(&ob, "Calls to add_message: %d   Packets: %d   Average packet size: %f\n\n",
	add_message_calls, inet_packets, (float) inet_volume / inet_packets);

#ifndef NO_ADD_ACTION
	stat_living_objects(&ob);
	outbuf_add(&ob, "\n");
#endif
#ifdef F_CACHE_STATS
	print_cache_stats(&ob);
	outbuf_add(&ob, "\n");
#endif
	print_swap_stats(&ob);
	outbuf_add(&ob, "\n");

        tot = show_otable_status(&ob, verbose);
        outbuf_add(&ob, "\n");
        tot += heart_beat_status(&ob, verbose);
        outbuf_add(&ob, "\n");
        tot += add_string_status(&ob, verbose);
        outbuf_add(&ob, "\n");
        tot += print_call_out_usage(&ob, verbose);
    } else {
	/* !verbose */
	outbuf_addv(&ob, "Sentences:\t\t\t%8d %8d\n", tot_alloc_sentence,
		    tot_alloc_sentence * sizeof(sentence_t));
	outbuf_addv(&ob, "Objects:\t\t\t%8d %8d\n",
		    tot_alloc_object, tot_alloc_object_size);
	outbuf_addv(&ob, "Prog blocks:\t\t\t%8d %8d\n",
		    total_num_prog_blocks, total_prog_block_size);
#ifdef ARRAY_STATS
	outbuf_addv(&ob, "Arrays:\t\t\t\t%8d %8d\n", num_arrays,
		    total_array_size);
#else
	outbuf_add(&ob, "<Array statistics disabled, no information available>\n");
#endif
	outbuf_addv(&ob, "Mappings:\t\t\t%8d %8d\n", num_mappings,
		    total_mapping_size);
	outbuf_addv(&ob, "Mappings(nodes):\t\t%8d\n", total_mapping_nodes);
	outbuf_addv(&ob, "Interactives:\t\t\t%8d %8d\n", total_users,
		    total_users * sizeof(interactive_t));

	tot = show_otable_status(&ob, verbose) +
	    heart_beat_status(&ob, verbose) +
	    add_string_status(&ob, verbose) +
	    print_call_out_usage(&ob, verbose);
    }

    tot += total_prog_block_size +
#ifdef ARRAY_STATS
	total_array_size +
#endif
	total_mapping_size +
	tot_alloc_sentence * sizeof(sentence_t) +
	tot_alloc_object_size +
	total_users * sizeof(interactive_t) +
	res;

    if (!verbose) {
	outbuf_add(&ob, "\t\t\t\t\t --------\n");
	outbuf_addv(&ob, "Total:\t\t\t\t\t %8d\n", tot);
    }
    outbuf_push(&ob);
}
#endif

#ifdef F_NOTIFY_FAIL
void
f_notify_fail(void)
{
    if (sp->type == T_STRING) {
	set_notify_fail_message(sp->u.string);
	free_string_svalue(sp--);
    }
    else {
	set_notify_fail_function(sp->u.fp);
	free_funp((sp--)->u.fp);
    }
}
#endif

#ifdef F_OBJECTP
void
f_objectp(void)
{
    if (sp->type == T_OBJECT){
        free_object(sp->u.ob, "f_objectp");
        *sp = const1;
    } else {
        free_svalue(sp, "f_objectp");
        *sp = const0;
    }
}
#endif

#ifdef F_OPCPROF
void
f_opcprof(void)
{
    if (st_num_arg == 1) {
        opcdump(sp->u.string);
        free_string_svalue(sp--);
    } else
	opcdump("/OPCPROF");
}
#endif

#ifdef F_ORIGIN
void
f_origin(void)
{
    push_constant_string(origin_name(caller_type));
}
#endif

#ifdef F_POINTERP
void
f_pointerp(void)
{
    if (sp->type == T_ARRAY){
        free_array(sp->u.arr);
        *sp = const1;
    } else {
        free_svalue(sp, "f_pointerp");
        *sp = const0;
    }
}
#endif

#ifdef F_PRESENT
void
f_present(void)
{
    int num_arg = st_num_arg;
    svalue_t *arg = sp - num_arg + 1;

#ifdef LAZY_RESETS
    if (num_arg == 2) {
	try_reset(arg[1].u.ob);
    }
#endif
    ob = object_present(arg, num_arg == 1 ? 0 : arg[1].u.ob);
    pop_n_elems(num_arg);
    if (ob) push_object(ob);
    else *++sp = const0;
}
#endif

#ifdef F_PREVIOUS_OBJECT
void
f_previous_object(void)
{
    control_stack_t *p;
    int i;

    if ((i = sp->u.number) > 0) {
        if (i >= CFG_MAX_CALL_DEPTH) {
            sp->u.number = 0;
            return;
	}
        ob = 0;
        p = csp;
        do {
            if ((p->framekind & FRAME_OB_CHANGE) && !(--i)) {
                ob = p->prev_ob;
                break;
	    }
	} while (--p >= control_stack);
    } else if (i == -1) {
        array_t *v;

        i = previous_ob ? 1 : 0;
        p = csp;
        do {
            if ((p->framekind & FRAME_OB_CHANGE) && p->prev_ob) i++;
	} while (--p >= control_stack);
        v = allocate_empty_array(i);
        p = csp;
        if (previous_ob){
	    if (!(previous_ob->flags & O_DESTRUCTED)) {
		v->item[0].type = T_OBJECT;
		v->item[0].u.ob = previous_ob;
		add_ref(previous_ob, "previous_object(-1)");
	    } else v->item[0] = const0;
            i = 1;
	} else i = 0;
        do {
            if ((p->framekind & FRAME_OB_CHANGE) && (ob = p->prev_ob)) {
		if (!(ob->flags & O_DESTRUCTED)){
		    v->item[i].type = T_OBJECT;
		    v->item[i].u.ob = ob;
		    add_ref(ob, "previous_object(-1)");
		} else v->item[i] = const0;
		i++;
	    }
	} while (--p >= control_stack);
        put_array(v);
        return;
    } else if (i < 0) {
        error("Illegal negative argument to previous_object()\n");
    } else
        ob = previous_ob;
    if (!ob || (ob->flags & O_DESTRUCTED))
        sp->u.number = 0;
    else {
        put_unrefed_undested_object(ob, "previous_object()");
    }
}
#endif

#ifdef F_PRINTF
void
f_printf(void)
{
    int num_arg = st_num_arg;
    char *ret;

    if (command_giver) {
	ret = string_print_formatted((sp - num_arg + 1)->u.string,
				     num_arg - 1, sp - num_arg + 2);
	if (ret) {
	    tell_object(command_giver, ret);
	    FREE_MSTR(ret);
	}
    }

    pop_n_elems(num_arg);
}
#endif

#ifdef F_PROCESS_STRING
void
f_process_string(void)
{
    char *str;

    str = process_string(sp->u.string);
    if (str != sp->u.string) {
        free_string_svalue(sp);
        put_malloced_string(str);
    }
}
#endif

#ifdef F_PROCESS_VALUE
void
f_process_value(void)
{
    svalue_t *ret;

    ret = process_value(sp->u.string);
    free_string_svalue(sp);
    if (ret) assign_svalue_no_free(sp, ret);
    else *sp = const0;
}
#endif

#ifdef F_QUERY_ED_MODE
void
f_query_ed_mode(void)
{
    /* n = prompt for line 'n'
       0 = normal ed prompt
       -1 = not in ed
       -2 = more prompt */
    if (current_object->flags & O_IN_EDIT) {
	push_number(object_ed_mode(current_object));
    } else
	push_number(-1);
}
#endif

#ifdef F_QUERY_HOST_NAME
void
f_query_host_name(void)
{
    char *tmp;

    if ((tmp = query_host_name()))
        push_constant_string(tmp);
    else
        push_number(0);
}
#endif

#ifdef F_QUERY_IDLE
void
f_query_idle(void)
{
    int i;

    i = query_idle(sp->u.ob);
    free_object(sp->u.ob, "f_query_idle");
    put_number(i);
}
#endif

#ifdef F_QUERY_IP_NAME
void
f_query_ip_name(void)
{
    char *tmp;

    tmp = query_ip_name(st_num_arg ? sp->u.ob : 0);
    if (st_num_arg) free_object((sp--)->u.ob, "f_query_ip_name");
    if (!tmp) *++sp = const0;
    else share_and_push_string(tmp);
}
#endif

#ifdef F_QUERY_IP_NUMBER
void
f_query_ip_number(void)
{
    char *tmp;

    tmp = query_ip_number(st_num_arg ? sp->u.ob : 0);
    if (st_num_arg) free_object((sp--)->u.ob, "f_query_ip_number");
    if (!tmp) *++sp = const0;
    else share_and_push_string(tmp);
}
#endif

#ifdef F_QUERY_LOAD_AVERAGE
void
f_query_load_average(void)
{
    copy_and_push_string(query_load_av());
}
#endif

#ifdef F_QUERY_PRIVS
void
f_query_privs(void)
{
    ob = sp->u.ob;
    if (ob->privs != NULL) {
        free_object(ob, "f_query_privs");
        sp->type = T_STRING;
        sp->u.string = make_shared_string(ob->privs);
        sp->subtype = STRING_SHARED;
    } else {
        free_object(ob, "f_query_privs");
        *sp = const0;
    }
}
#endif

#ifdef F_QUERY_SNOOPING
void
f_query_snooping(void)
{
    ob = query_snooping(sp->u.ob);
    free_object(sp->u.ob, "f_query_snooping");
    if (ob) { put_unrefed_undested_object(ob, "query_snooping"); }
    else *sp = const0;
}
#endif

#ifdef F_QUERY_SNOOP
void
f_query_snoop(void)
{
    ob = query_snoop(sp->u.ob);
    free_object(sp->u.ob, "f_query_snoop");
    if (ob) { put_unrefed_undested_object(ob, "query_snoop"); }
    else *sp = const0;
}
#endif

#ifdef F_QUERY_VERB
void
f_query_verb(void)
{
    if (!last_verb) {
        push_number(0);
        return;
    }
    share_and_push_string(last_verb);
}
#endif

#ifdef F_RANDOM
void
f_random(void)
{
    if (sp->u.number <= 0) {
	sp->u.number = 0;
	return;
    }
    sp->u.number = rand() % sp->u.number;
}
#endif

#ifdef F_READ_BYTES
void
f_read_bytes(void)
{
    char *str;
    int start = 0, len = 0, rlen = 0, num_arg = st_num_arg;
    svalue_t *arg;

    arg = sp - num_arg + 1;
    if (num_arg > 1)
        start = arg[1].u.number;
    if (num_arg == 3) {
        len = arg[2].u.number;
    }
    str = read_bytes(arg[0].u.string, start, len, &rlen);
    pop_n_elems(num_arg);
    if (str == 0)
        push_number(0);
    else {
        push_malloced_string(str);
    }
}
#endif

#ifdef F_READ_BUFFER
void
f_read_buffer(void)
{
    char *str;
    int start = 0, len = 0, rlen = 0, num_arg = st_num_arg;
    int from_file = 0;          /* new line */
    svalue_t *arg = sp - num_arg + 1;

    if (num_arg > 1) {
       start = arg[1].u.number;
       if (num_arg == 3) {
	   len = arg[2].u.number;
       }
    }
    if (arg[0].type == T_STRING) {
        from_file = 1;          /* new line */
        str = read_bytes(arg[0].u.string, start, len, &rlen);
    } else {                    /* T_BUFFER */
        str = read_buffer(arg[0].u.buf, start, len, &rlen);
    }
    pop_n_elems(num_arg);
    if (str == 0) {
        push_number(0);
    } else if (from_file) {     /* changed */
        buffer_t *buf;

        buf = allocate_buffer(rlen);
        memcpy(buf->item, str, rlen);
        (++sp)->u.buf = buf;
        sp->type = T_BUFFER;
        FREE_MSTR(str);
    } else {                    /* T_BUFFER */
        push_malloced_string(str);
    }
}
#endif

#ifdef F_READ_FILE
void
f_read_file(void)
{
    char *str;
    int start,len;

    if (st_num_arg == 3) {
        len = (sp--)->u.number;
    } else len = 0;
    if (st_num_arg > 1)
        start = (sp--)->u.number;
    else start = 0;

    str = read_file(sp->u.string, start, len);
    free_string_svalue(sp);
    if (!str) *sp = const0;
    else { sp->subtype = STRING_MALLOC; sp->u.string = str; }
}
#endif

#ifdef F_RECEIVE
void
f_receive(void)
{
    if (current_object->interactive) {
	check_legal_string(sp->u.string);
	add_message(current_object, sp->u.string);
    }
    free_string_svalue(sp--);
}
#endif

#ifdef F_REG_ASSOC
void
f_reg_assoc(void) {
    svalue_t *arg;
    array_t *vec;

    arg = sp - st_num_arg + 1;

    if (!(arg[2].type == T_ARRAY))
	error("Bad argument 3 to reg_assoc()\n");

    vec = reg_assoc(arg[0].u.string, arg[1].u.arr, arg[2].u.arr, st_num_arg > 3 ? &arg[3] : &const0);

    if (st_num_arg == 4)
	pop_3_elems();
    else
	pop_2_elems();
    free_string_svalue(sp);
    sp->type = T_ARRAY;
    sp->u.arr = vec;
}
#endif

#ifdef F_REGEXP
void
f_regexp(void)
{
    array_t *v;
    int flag;

    if (st_num_arg > 2) {
        if (!(sp->type == T_NUMBER)) error("Bad argument 3 to regexp()\n");
	if (sp[-2].type == T_STRING) error("3rd argument illegal for regexp(string, string)\n");
	flag = (sp--)->u.number;
    } else flag = 0;
    if (sp[-1].type == T_STRING) {
	flag = match_single_regexp((sp - 1)->u.string, sp->u.string);
	free_string_svalue(sp--);
	free_string_svalue(sp);
	put_number(flag);
    } else {
	v = match_regexp((sp - 1)->u.arr, sp->u.string, flag);

	free_string_svalue(sp--);
	free_array(sp->u.arr);
	sp->u.arr = v;
    }
}
#endif

#ifdef F_REMOVE_ACTION
void
f_remove_action(void)
{
    int success;

    success = remove_action((sp - 1)->u.string, sp->u.string);
    free_string_svalue(sp--);
    free_string_svalue(sp);
    put_number(success);
}
#endif

#ifdef F_REMOVE_CALL_OUT
void
f_remove_call_out(void)
{
    int i;

    if (st_num_arg) {
#ifdef CALLOUT_HANDLES
	if (sp->type == T_STRING) {
#endif
	    i = remove_call_out(current_object, sp->u.string);
	    free_string_svalue(sp);
#ifdef CALLOUT_HANDLES
	} else {
	    i = remove_call_out_by_handle(sp->u.number);
	}
#endif
    } else {
	remove_all_call_out(current_object);
	i = 0;
	sp++;
    }
    put_number(i);
}
#endif

#ifdef F_RENAME
void
f_rename(void)
{
    int i;

    i = do_rename((sp - 1)->u.string, sp->u.string, F_RENAME);
    free_string_svalue(sp--);
    free_string_svalue(sp);
    put_number(i);
}
#endif				/* F_RENAME */

/* This is an enhancement to the f_replace_string() in efuns_main.c of
   MudOS v21.  When the search pattern has more than one character,
   this version of f_replace_string() uses a skip table to more efficiently
   search the file for the search pattern (the basic idea is to avoid
   strings comparisons where possible).  This version is anywhere from
   15% to 40% faster than the old version depending on the size of the
   string to be searched and the length of the search string (and depending
   on the relative frequency with which the letters in the search string
   appear in the string to be searched).

   Note: this version should behave identically to the old version (except
   for runtime).  When the search pattern is only one character long, the
   old algorithm is used.  The new algorithm is actually about 10% slower
   than the old one when the search string is only one character long.

   This enhancement to f_replace_string() was written by John Garnett
   (aka Truilkan) on 1995/04/29.  I believe the original replace_string()
   was written by Dave Richards (Cygnus).

   I didn't come up with the idea of this algorithm (learned it in
   a university programming course way back when).  For those interested
   in the workings of the algorithm, you can probably find it in a book on
   string processing algorithms.  Its also fairly easy to figure out the
   algorithm by tracing through it for a small example.
*/
#ifdef F_REPLACE_STRING

/*
syntax for replace_string is now:
    string replace_string(src, pat, rep);   // or
    string replace_string(src, pat, rep, max);  // or
    string replace_string(src, pat, rep, first, last);

The 4th/5th args are optional (to retain backward compatibility).
- src, pat, and rep are all strings.
- max is an integer. It will replace all occurances up to max
  matches (starting as 1 as the first), with a value of 0 meaning
  'replace all')
- first and last are just a range to replace between, with
  the following constraints
    first < 1: change all from start
    last == 0 || last > max matches:    change all to end
    first > last: return unmodified array.
(i.e, with 4 args, it's like calling it with:
    replace_string(src, pat, rep, 0, max);
)
*/

void
f_replace_string(void)
{
    int plen, rlen, dlen, slen, first, last, cur, j;

    char *pattern;
    char *replace;
    register char *src, *dst1, *dst2;
    svalue_t *arg;
    int skip_table[256];
    char *slimit = NULL;
    char *flimit = NULL;
    char *climit = NULL;
    int probe = 0;
    int skip;

    if (st_num_arg > 5) {
        error("Too many args to replace_string.\n");
        pop_n_elems(st_num_arg);
        return;
    }
    arg = sp - st_num_arg + 1;
    src = arg->u.string;
    first = 0;
    last = 0;

    /* ˬdǤJ replace_string ѼƭӼ */
    if (st_num_arg >= 4) {
	CHECK_TYPES((arg+3), T_NUMBER, 4, F_REPLACE_STRING);
	first = (arg+3)->u.number;

	if (st_num_arg == 4) {
	    last = first;
	    first = 0;
        } else if (st_num_arg == 5) {
	    CHECK_TYPES((arg+4), T_NUMBER, 5, F_REPLACE_STRING);
	    /* first was set above. */
	    last = sp->u.number;
	}
    }

    if (!last) last = max_string_length;

    if (first > last) {         /* just return it */
        pop_n_elems(st_num_arg - 1);
        return;
    }

    pattern = (arg+1)->u.string;
    plen = SVALUE_STRLEN(arg+1);
    if ( plen < 1 ) {
        pop_n_elems(st_num_arg - 1);    /* just return it */
        return;
    }

    replace = (arg+2)->u.string;
    rlen = SVALUE_STRLEN(arg+2);
    dlen = 0;
    cur = 0;

    if (rlen <= plen) {
	/* replace  pattern u -> rפiAݭs
	 * tmO */
	unlink_string_svalue(arg);
	src = arg->u.string;
    }

    if (plen > 1) {
	/* build skip table */
	for (j = 0; j < 256; j++) skip_table[j] = plen;
	for (j = 0; j < plen; j++)
	    skip_table[(unsigned char)pattern[j]] = plen - j - 1;

	slen = SVALUE_STRLEN(arg);
	slimit = src + slen;
	flimit = slimit - plen + 1;
	probe = plen - 1;
    }

    if (rlen <= plen) {
        /* rꤺN覡 */
        dst2 = dst1 = arg->u.string;

        if (plen > 1) {
	    /* pattern פj 1, ϥ jump table t׸ */
            while (src < flimit) {
                if ((skip = skip_table[(unsigned char)src[probe]])) {
                    for (climit = dst2 + skip; dst2 < climit; *dst2++ = *src++)
                        ;
                } else if (memcmp(src, pattern, plen) == 0) {
                    cur++;
                    if ((cur >= first) &&  (cur <= last)) {
                        if (rlen) {
                            memcpy(dst2, replace, rlen);
                            dst2 += rlen;
                        }
                        src += plen;
                        if (cur == last) break;
                    } else {
			memcpy(dst2, src, plen);
			dst2 += plen;
                        src += plen;
                    }
                } else {
                    *dst2++ = *src++;
                }
            }
            memcpy(dst2, src, slimit - src);
            dst2 += (slimit - src);
	    *dst2 = 0;
	    arg->u.string = extend_string(dst1, dst2 - dst1);
	} else {
	    /* pattern  == 1, ZOķ */
	    if( rlen ) {	/* ] rlen<=plen, B plen==1, ҥH rlen==1 */
		while (*src) {
		    if (*src == *pattern) {
			cur++;
			if( cur < first ) continue;
			*src = *replace;
			if( cur > last ) break;
		    }
		    src++;
		}
	    } else { /* rlen is zero */
		while (*src) {
		    if (*src++ == *pattern) {
			cur++;
			if (cur >= first) {
			    dst2 = src - 1;
			    while (*src) {
				if (*src == *pattern) {
				    cur++;
				    if (cur <= last) { src++; continue; }
				    else {
					while (*src)
					    *dst2++ = *src++;
					break;
				    }
				}
				*dst2++ = *src++;
			    }
			    *dst2 = '\0';
			    arg->u.string = extend_string(dst1, dst2 - dst1);
			    break;
			}
		    }
		}
	    }
	}
        pop_n_elems(st_num_arg - 1);
    }
    else
    {
        dst2 = dst1 = new_string(max_string_length, "f_replace_string: 2");

        if (plen > 1) {
            while (src < flimit) {
                if ((skip = skip_table[(unsigned char)src[probe]])) {
                    for (climit = dst2 + skip; dst2 < climit; *dst2++ = *src++)
                        ;

                } else if (memcmp(src, pattern, plen) == 0) {
                    cur++;
                    if ((cur >= first) && (cur <= last)) {
			if (max_string_length - dlen <= rlen) {
			    pop_n_elems(st_num_arg);
			    push_svalue(&const0u);
			    FREE_MSTR(dst1);
			    return;
			}
			memcpy(dst2, replace, rlen);
			dst2 += rlen;
			dlen += rlen;
			src += plen;
                        if (cur == last) break;
                    } else {
			dlen += plen;
			if (max_string_length - dlen <= 0) {
			    pop_n_elems(st_num_arg);
			    push_svalue(&const0u);

			    FREE_MSTR(dst1);
			    return;
			}
			memcpy(dst2, src, plen);
			dst2 += plen;
                        src += plen;
                    }
                } else {
                    if (max_string_length - dlen <= 1) {
                        pop_n_elems(st_num_arg);
                        push_svalue(&const0u);

                        FREE_MSTR(dst1);
                        return;
                    }
                    *dst2++ = *src++;
                    dlen++;
                }
            }
            if (max_string_length - dlen <= (slimit - src)) {
                pop_n_elems(st_num_arg);
                push_svalue(&const0u);
                FREE_MSTR(dst1);
                return;
            }
            memcpy(dst2, src, slimit - src);
            dst2 += (slimit - src);
        } else { /* plen <= 1 */
	    /* Beek: plen == 1 */
            while (*src != '\0') {
                if (*src == *pattern) {
                    cur++;
                    if (cur >= first && cur <= last) {
                        if (rlen != 0) {
                            if (max_string_length - dlen <= rlen) {
                                pop_n_elems(st_num_arg);
                                push_svalue(&const0u);
                                FREE_MSTR(dst1);
                                return;
                            }
                            strncpy(dst2, replace, rlen);
                            dst2 += rlen;
                            dlen += rlen;
                        }
                        src++;
                        continue;
                    }
                }
                if (max_string_length - dlen <= 1) {
                    pop_n_elems(st_num_arg);
                    push_svalue(&const0u);
                    FREE_MSTR(dst1);
                    return;
                }
                *dst2++ = *src++;
                dlen++;
            }
        }
        *dst2 = '\0';

        pop_n_elems(st_num_arg);
        /*
         * shrink block or make a copy of exact size
         */
        push_malloced_string(extend_string(dst1, dst2 - dst1));
    }
}
#endif

#ifdef F_RESOLVE
void
f_resolve(void)
{
    int i, query_addr_number(char *, char *);

    i = query_addr_number((sp - 1)->u.string, sp->u.string);
    free_string_svalue(sp--);
    free_string_svalue(sp);
    put_number(i);
}				/* f_resolve() */
#endif

#ifdef F_RESTORE_OBJECT
void
f_restore_object(void)
{
    int flag;

    flag = (st_num_arg > 1) ? (sp--)->u.number : 0;
    flag = restore_object(current_object, sp->u.string, flag);
    free_string_svalue(sp);
    put_number(flag);
}
#endif

#ifdef F_RESTORE_VARIABLE
void
f_restore_variable(void) {
    svalue_t v;

    unlink_string_svalue(sp);
    v.type = T_NUMBER;

    restore_variable(&v, sp->u.string);
    FREE_MSTR(sp->u.string);
    *sp = v;
}
#endif

#ifdef F_RM
void
f_rm(void)
{
    int i;

    i = remove_file(sp->u.string);
    free_string_svalue(sp);
    put_number(i);
}
#endif

#ifdef F_RMDIR
void
f_rmdir(void)
{
    char *path;

    path = check_valid_path(sp->u.string, current_object, "rmdir", 1);
    if (!path || rmdir(path) == -1){
        free_string_svalue(sp);
        *sp = const0;
    }
    else {
        free_string_svalue(sp);
        *sp = const1;
    }
}
#endif

#ifdef F_SAVE_OBJECT
void
f_save_object(void)
{
    int flag;

    flag = (st_num_arg == 2) && (sp--)->u.number;
    flag = save_object(current_object, sp->u.string, flag);
    free_string_svalue(sp);
    put_number(flag);
}
#endif

#ifdef F_SAVE_VARIABLE
void
f_save_variable(void) {
    char *p;

    p = save_variable(sp);
    pop_stack();
    push_malloced_string(p);
}
#endif

#ifdef F_SAY
void
f_say(void)
{
    array_t *avoid;
    static array_t vtmp =
    {1,
     1,
    };

    if (st_num_arg == 1) {
	avoid = &the_null_array;
	say(sp, avoid);
	pop_stack();
    } else {
	if (sp->type == T_OBJECT) {
	    vtmp.item[0].type = T_OBJECT;
	    vtmp.item[0].u.ob = sp->u.ob;
	    avoid = &vtmp;
	} else {		/* must be a array... */
	    avoid = sp->u.arr;
	}
	say(sp - 1, avoid);
	pop_2_elems();
    }
}
#endif

#ifdef F_SET_EVAL_LIMIT
/* warning: do not enable this without using valid_override() in the master
   object and a set_eval_limit() simul_efun to restrict access.
*/
void
f_set_eval_limit(void)
{
    switch (sp->u.number) {
    case 0:
	sp->u.number = eval_cost = max_cost;
	break;
    case -1:
	sp->u.number = eval_cost;
	break;
    case 1:
	sp->u.number = max_cost;
	break;
    default:
	max_cost = sp->u.number;
	break;
    }
}
#endif

#ifdef F_SET_BIT
void
f_set_bit(void)
{
    char *str;
    int len, old_len, ind, bit;

    if (sp->u.number > MAX_BITS)
        error("set_bit() bit requested: %d > maximum bits: %d\n", sp->u.number, MAX_BITS);
    bit = (sp--)->u.number;
    if (bit < 0)
	error("Bad argument 2 (negative) to set_bit().\n");
    ind = bit/6;
    bit %= 6;
    old_len = len = SVALUE_STRLEN(sp);
    if (ind >= len)
        len = ind + 1;
    if (ind < old_len) {
	unlink_string_svalue(sp);
	str = sp->u.string;
    } else {
	str = new_string(len, "f_set_bit: str");
	str[len] = '\0';
	if (old_len)
	    memcpy(str, sp->u.string, old_len);
	if (len > old_len)
	    memset(str + old_len, ' ', len - old_len);
	free_string_svalue(sp);
	sp->subtype = STRING_MALLOC;
	sp->u.string = str;
    }

    if (str[ind] > 0x3f + ' ' || str[ind] < ' ')
        error("Illegal bit pattern in set_bit character %d\n", ind);
    str[ind] = ((str[ind] - ' ') | (1 << bit)) + ' ';
}
#endif

#ifdef F_SET_HEART_BEAT
void
f_set_heart_beat(void)
{
    set_heart_beat(current_object, (sp--)->u.number);
}
#endif

#ifdef F_QUERY_HEART_BEAT
void
f_query_heart_beat(void)
{
    free_object(ob = sp->u.ob, "f_query_heart_beat");
    put_number(query_heart_beat(ob));
}
#endif

#ifdef F_SET_HIDE
void
f_set_hide(void)
{
    if (!valid_hide(current_object)) {
	sp--;
	return;
    }
    if ((sp--)->u.number) {
	if (!(current_object->flags & O_HIDDEN) && current_object->interactive)
	    num_hidden++;
	current_object->flags |= O_HIDDEN;
    } else {
	if ((current_object->flags & O_HIDDEN) && current_object->interactive)
	    num_hidden--;
	current_object->flags &= ~O_HIDDEN;
    }
}
#endif

#ifdef F_SET_LIGHT
void
f_set_light(void)
{
    object_t *o1;

    add_light(current_object, sp->u.number);
    o1 = current_object;
#ifndef NO_ENVIRONMENT
    while (o1->super)
	o1 = o1->super;
#endif
    sp->u.number = o1->total_light;
}
#endif

#ifdef F_SET_LIVING_NAME
void
f_set_living_name(void)
{
    set_living_name(current_object, sp->u.string);
    free_string_svalue(sp--);
}
#endif

#ifdef F_SET_PRIVS
void
f_set_privs(void)
{
    object_t *ob;

    ob = (sp - 1)->u.ob;
    if (ob->privs != NULL)
        free_string(ob->privs);
    if (!(sp->type == T_STRING)) {
        ob->privs = NULL;
	sp--; /* It's a number */
    } else {
        ob->privs = make_shared_string(sp->u.string);
	free_string_svalue(sp--);
    }
    free_object(ob, "f_set_privs");
    sp--;
}
#endif

#ifdef F_SHADOW
void
f_shadow(void)
{
    object_t *ob;

    ob = (sp - 1)->u.ob;
    if (!((sp--)->u.number)) {
        ob = ob->shadowed;
        free_object(sp->u.ob, "f_shadow:1");
        if (ob) {
	  add_ref(ob, "shadow(ob, 0)");
	  sp->u.ob = ob;
	}
        else *sp = const0;
        return;
    }
    if (ob == current_object) {
        error("shadow: Can't shadow self\n");
    }
    if (validate_shadowing(ob)) {
        if (current_object->flags & O_DESTRUCTED) {
            free_object(ob, "f_shadow:2");
            *sp = const0;
            return;
	}
        /*
         * The shadow is entered first in the chain.
         */
        while (ob->shadowed)
            ob = ob->shadowed;
        current_object->shadowing = ob;
        ob->shadowed = current_object;
        free_object(sp->u.ob, "f_shadow:3");
	add_ref(ob, "shadow(ob, 1)");
        sp->u.ob = ob;
        return;
    }
    free_object(sp->u.ob, "f_shadow:4");
    *sp = const0;
}
#endif

#ifdef F_SHOUT
void
f_shout(void)
{
    shout_string(sp->u.string);
    free_string_svalue(sp--);
}
#endif

#ifdef F_SHUTDOWN
void
f_shutdown(void)
{
    shutdownMudOS(st_num_arg ? sp->u.number : (*++sp = const0, 0));
}
#endif

#ifdef F_SIZEOF
void
f_sizeof(void)
{
    int i;

    switch (sp->type) {
    case T_CLASS:
	i = sp->u.arr->size;
	free_class(sp->u.arr);
	break;
    case T_ARRAY:
	i = sp->u.arr->size;
	free_array(sp->u.arr);
	break;
    case T_MAPPING:
	i = sp->u.map->count;
	free_mapping(sp->u.map);
	break;
    case T_BUFFER:
	i = sp->u.buf->size;
	free_buffer(sp->u.buf);
	break;
    case T_STRING:
	i = SVALUE_STRLEN(sp);
	free_string_svalue(sp);
	break;
    default:
	i = 0;
	free_svalue(sp, "f_sizeof");
    }
    sp->type = T_NUMBER;
    sp->u.number = i;
}
#endif

#ifdef F_SNOOP
void
f_snoop(void)
{
    /*
     * This one takes a variable number of arguments. It returns 0 or an
     * object.
     */
    if (st_num_arg == 1) {
        if (!new_set_snoop(sp->u.ob, 0) || (sp->u.ob->flags & O_DESTRUCTED)) {
	    free_object(sp->u.ob, "f_snoop:1");
	    *sp = const0;
	}
    } else {
        if (!new_set_snoop((sp - 1)->u.ob, sp->u.ob) ||
	    (sp->u.ob->flags & O_DESTRUCTED)) {
	    free_object((sp--)->u.ob, "f_snoop:2");
	    free_object(sp->u.ob, "f_snoop:3");
	    *sp = const0;
	} else {
	    free_object((--sp)->u.ob, "f_snoop:4");
	    sp->u.ob = (sp+1)->u.ob;
	}
    }
}
#endif

#ifdef F_SPRINTF
void
f_sprintf(void)
{
    char *s;
    int num_arg = st_num_arg;

    s = string_print_formatted((sp - num_arg + 1)->u.string,
                               num_arg - 1, sp - num_arg + 2);
    pop_n_elems(num_arg);

    (++sp)->type = T_STRING;
    if (!s) {
	sp->subtype = STRING_CONSTANT;
	sp->u.string = "";
    } else {
	sp->subtype = STRING_MALLOC;
	sp->u.string = s;
    }
}
#endif

#ifdef F_STAT
void
f_stat(void)
{
    struct stat buf;
    char *path;
    array_t *v;

    path = check_valid_path((--sp)->u.string, current_object, "stat", 0);
    if (!path) {
        free_string_svalue(sp);
        *sp = const0;
        return;
    }
    if (stat(path, &buf) != -1) {
        if (buf.st_mode & S_IFREG) {    /* if a regular file */
            v = allocate_empty_array(3);
            v->item[0].type = T_NUMBER;
            v->item[0].u.number = buf.st_size;
            v->item[1].type = T_NUMBER;
            v->item[1].u.number = buf.st_mtime;
            v->item[2].type = T_NUMBER;
            ob = find_object2(path);
            if (ob && !object_visible(ob))
                ob = 0;
            if (ob)
                v->item[2].u.number = ob->load_time;
            else
                v->item[2].u.number = 0;
            free_string_svalue(sp);
            put_array(v);
            return;
	}
    }
    v = get_dir(sp->u.string, (sp+1)->u.number);
    free_string_svalue(sp);
    if (v) { put_array(v); }
    else *sp = const0;
}
#endif

#ifdef F_STRSRCH
/*
 * int strsrch(string big, string little, [ int flag ])
 * - search for little in big, starting at right if flag is set
 *   return int offset of little, -1 if not found
 *
 * Written 930706 by Luke Mewburn <zak@rmit.edu.au>
 */

void
f_strsrch(void)
{
    register char *big, *little, *pos;
    static char buf[2];         /* should be initialized to 0 */
    int i, blen, llen;

    sp--;
    big = (sp - 1)->u.string;
    blen = SVALUE_STRLEN(sp - 1);
    if (sp->type == T_NUMBER) {
        little = buf;
        if ((buf[0] = (char) sp->u.number))
	    llen = 1;
	else
	    llen = 0;
    } else {
        little = sp->u.string;
	llen = SVALUE_STRLEN(sp);
    }

    if (!llen || blen < llen) {
        pos = NULL;

        /* start at left */
    } else if (!((sp+1)->u.number)) {
        if (!little[1])         /* 1 char srch pattern */
            pos = strchr(big, (int) little[0]);
        else
            pos = (char *)strstr(big, little);
        /* start at right */
    } else {                    /* XXX: maybe test for -1 */
        if (!little[1])         /* 1 char srch pattern */
            pos = strrchr(big, (int) little[0]);
        else {
            char c = *little;

            pos = big + blen;   /* find end */
            pos -= llen;        /* find rightmost pos it _can_ be */
            do {
                do {
                    if (*pos == c) break;
		} while (--pos >= big);
                if (*pos != c) {
                    pos = NULL;
                    break;
		}
                for (i = 1; little[i] && (pos[i] == little[i]); i++);   /* scan all chars */
                if (!little[i])
                    break;
	    } while (--pos >= big);
	}
    }

    if (!pos)
        i = -1;
    else
        i = (int) (pos - big);
    if (sp->type == T_STRING) free_string_svalue(sp);
    free_string_svalue(--sp);
    put_number(i);
}				/* strsrch */
#endif

#ifdef F_STRCMP
void
f_strcmp(void)
{
    int i;

    i = strcmp((sp - 1)->u.string, sp->u.string);
    free_string_svalue(sp--);
    free_string_svalue(sp);
    put_number(i);
}
#endif

#ifdef F_STRINGP
void
f_stringp(void)
{
    if (sp->type == T_STRING){
        free_string_svalue(sp);
        *sp = const1;
    }
    else {
        free_svalue(sp, "f_stringp");
        *sp = const0;
    }
}
#endif

#ifdef F_BUFFERP
void
f_bufferp(void)
{
    if (sp->type == T_BUFFER) {
        free_buffer(sp->u.buf);
        *sp = const1;
    } else {
        free_svalue(sp, "f_bufferp");
        *sp = const0;
    }
}
#endif

#ifdef F_SWAP
void
f_swap(void)
{
    object_t *ob = sp->u.ob;
    control_stack_t *p;

    /* a few sanity checks */
    if (!(ob->flags & O_SWAPPED) && (ob != current_object)) {
	for (p = csp; p >= control_stack; p--)
	    if (ob == csp->ob) {
		pop_stack();
		return;
	    }
	(void) swap(sp->u.ob);
    }

    pop_stack();
}
#endif

#ifdef F_TAIL
void
f_tail(void)
{
    if (tail(sp->u.string)){
        free_string_svalue(sp);
        *sp = const1;
    }
    else {
	free_string_svalue(sp);
	*sp = const0;
    }
}
#endif

#ifdef F_TELL_OBJECT
void
f_tell_object(void)
{
    tell_object((sp - 1)->u.ob, sp->u.string);
    free_string_svalue(sp--);
    pop_stack();
}
#endif

#ifdef F_TELL_ROOM
void
f_tell_room(void)
{
    array_t *avoid;
    int num_arg = st_num_arg;
    svalue_t *arg = sp - num_arg + 1;

    if (arg->type == T_OBJECT) {
        ob = arg[0].u.ob;
    } else {                    /* must be a string... */
        ob = find_object(arg[0].u.string);
        if (!ob || !object_visible(ob))
            error("Bad argument 1 to tell_room()\n");
    }

    if (num_arg == 2) {
        avoid = &the_null_array;
    } else {
        avoid = arg[2].u.arr;
    }

    tell_room(ob, &arg[1], avoid);
    free_array(avoid);
    free_svalue(arg + 1, "f_tell_room");
    free_svalue(arg, "f_tell_room");
    sp = arg - 1;
}
#endif

#ifdef F_TEST_BIT
void
f_test_bit(void)
{
    int ind = (sp--)->u.number;

    if (ind / 6 >= SVALUE_STRLEN(sp)) {
        free_string_svalue(sp);
        *sp = const0;
        return;
    }
    if (ind < 0) error("Bad argument 2 (negative) to test_bit().\n");
    if ((sp->u.string[ind / 6] - ' ') & (1 << (ind % 6))) {
        free_string_svalue(sp);
        *sp = const1;
    } else {
        free_string_svalue(sp);
        *sp = const0;
    }
}
#endif

#ifdef F_NEXT_BIT
void
f_next_bit(void)
{
    int start = (sp--)->u.number;
    int len = SVALUE_STRLEN(sp);
    int which, bit = 0, value;

    if (!len || start / 6 >= len) {
        free_string_svalue(sp);
	put_number(-1);
        return;
    }
    /* Find the next bit AFTER start */
    if (start > 0) {
	if (start % 6 == 5) {
	    which = (start / 6) + 1;
	    value = sp->u.string[which] - ' ';
	} else {
	    /* we have a partial byte to check */
	    which = start / 6;
	    bit = 0x3f - ((1 << ((start % 6) + 1)) - 1);
	    value = (sp->u.string[which] - ' ') & bit;
	}
    } else {
	which = 0;
	value = *sp->u.string - ' ';
    }

    while (1) {
	if (value)  {
	    if (value & 0x07) {
		if (value & 0x01)
		    bit = which * 6;
		else if (value & 0x02)
		    bit = which * 6 + 1;
		else if (value & 0x04)
		    bit = which * 6 + 2;
		break;
	    } else if (value & 0x38) {
		if (value & 0x08)
		    bit = which * 6 + 3;
		else if (value & 0x10)
		    bit = which * 6 + 4;
		else if (value & 0x20)
		    bit = which * 6 + 5;
		break;
	    }
	}
	which++;
	if (which == len) {
	    bit = -1;
	    break;
	}
	value = sp->u.string[which] - ' ';
    }

    free_string_svalue(sp);
    put_number(bit);
}
#endif

#ifdef F_THIS_OBJECT
void
f_this_object(void)
{
    if (current_object->flags & O_DESTRUCTED)   /* Fixed from 3.1.1 */
        *++sp = const0;
    else
        push_object(current_object);
}
#endif

#ifdef F_THIS_PLAYER
void
f_this_player(void)
{
    if (sp->u.number) {
	if (current_interactive)
	    put_unrefed_object(current_interactive, "this_player(1)");
	else sp->u.number = 0;
    } else {
	if (command_giver)
	    put_unrefed_object(command_giver, "this_player(0)");
	/* else zero is on stack already */
    }
}
#endif

#ifdef F_SET_THIS_PLAYER
void
f_set_this_player(void)
{
    if (sp->type == T_NUMBER)
	command_giver = 0;
    else
	command_giver = sp->u.ob;
    pop_stack();
}
#endif

#ifdef F_THROW
void
f_throw(void)
{
    free_svalue(&catch_value, "f_throw");
    catch_value = *sp--;
    throw_error();		/* do the longjump, with extra checks... */
}
#endif

#ifdef F_TIME
void
f_time(void)
{
#ifdef WIN32
    /* Necessary?  Good? -Beek */
    push_number(current_time = time(NULL));
#else
    push_number(time(NULL));
#endif
}
#endif

#ifdef F_TO_FLOAT
void
f_to_float(void)
{
    double temp = 0;

    switch(sp->type){
        case T_NUMBER:
            sp->type = T_REAL;
            sp->u.real = (double) sp->u.number;
            break;
        case T_STRING:
            sscanf(sp->u.string, "%lf", &temp);
            free_string_svalue(sp);
            sp->type = T_REAL;
            sp->u.real = temp;
    }
}
#endif

#ifdef F_TO_INT
void
f_to_int(void)
{
    switch(sp->type){
        case T_REAL:
            sp->type = T_NUMBER;
            sp->u.number = (int) sp->u.real;
            break;
        case T_STRING:
        {
            int temp;

            temp = atoi(sp->u.string);
            free_string_svalue(sp);
            sp->u.number = temp;
            sp->type = T_NUMBER;
            break;
	}
        case T_BUFFER:
            if (sp->u.buf->size < sizeof(int)) {
                free_buffer(sp->u.buf);
                *sp = const0;
	    } else {
                int hostint, netint;

                memcpy((char *) &netint, sp->u.buf->item, sizeof(int));
                hostint = ntohl(netint);
                free_buffer(sp->u.buf);
                put_number(hostint);
	    }
    }
}
#endif

#ifdef F_TYPEOF
void
f_typeof(void)
{
    char *t = type_name(sp->type);

    free_svalue(sp, "f_typeof");
    put_constant_string(t);
}
#endif

#ifdef F_UNDEFINEDP
void
f_undefinedp(void)
{
    if (sp->type == T_NUMBER){
        if (!sp->u.number && (sp->subtype == T_UNDEFINED)) {
	    *sp = const1;
        } else *sp = const0;
    } else {
        free_svalue(sp, "f_undefinedp");
        *sp = const0;
    }
}
#endif

#ifdef F_UPTIME
void
f_uptime(void)
{
    push_number(time(NULL) - boot_time);
}
#endif

#ifdef F_USERP
void
f_userp(void)
{
    int i;

    i = (int) sp->u.ob->flags & O_ONCE_INTERACTIVE;
    free_object(sp->u.ob, "f_userp");
    put_number(i != 0);
}
#endif

#ifdef F_USERS
void
f_users(void)
{
    push_refed_array(users());
}
#endif

#ifdef F_WIZARDP
void
f_wizardp(void)
{
    int i;

    i = (int) sp->u.ob->flags & O_IS_WIZARD;
    free_object(sp->u.ob, "f_wizardp");
    put_number(i != 0);
}
#endif

#ifdef F_VIRTUALP
void
f_virtualp(void)
{
    int i;

    i = (int) sp->u.ob->flags & O_VIRTUAL;
    free_object(sp->u.ob, "f_virtualp");
    put_number(i != 0);
}
#endif

#ifdef F_WRITE
void
f_write(void)
{
    do_write(sp);
    pop_stack();
}
#endif

#ifdef F_WRITE_BYTES
void
f_write_bytes(void)
{
    int i = 0;

    switch(sp->type){
        case T_NUMBER:
        {
            int netint;
            char *netbuf;

            if (!sp->u.number) bad_arg(3, F_WRITE_BYTES);
            netint = htonl(sp->u.number);       /* convert to network
                                                 * byte-order */
            netbuf = (char *) &netint;
            i = write_bytes((sp - 2)->u.string, (sp - 1)->u.number, netbuf,
                            sizeof(int));
            break;
	}

        case T_BUFFER:
            i = write_bytes((sp - 2)->u.string, (sp - 1)->u.number,
                            (char *) sp->u.buf->item, sp->u.buf->size);
            break;
        case T_STRING:
            i = write_bytes((sp - 2)->u.string, (sp - 1)->u.number,
                            sp->u.string, SVALUE_STRLEN(sp));
            break;
        default:
            bad_argument(sp, T_BUFFER|T_STRING|T_NUMBER, 3, F_WRITE_BYTES);
    }
    free_svalue(sp--, "f_write_bytes");
    free_string_svalue(--sp);
    put_number(i);
}
#endif

#ifdef F_WRITE_BUFFER
void
f_write_buffer(void)
{
    int i = 0;

    if ((sp-2)->type == T_STRING){
        f_write_bytes();
        return;
    }

    switch(sp->type){
        case T_NUMBER:
        {
            int netint;
            char *netbuf;

            netint = htonl(sp->u.number);       /* convert to network
                                                 * byte-order */
            netbuf = (char *) &netint;
            i = write_buffer((sp - 2)->u.buf, (sp - 1)->u.number, netbuf,
                            sizeof(int));
            break;
	}

        case T_BUFFER:
            i = write_buffer((sp - 2)->u.buf, (sp - 1)->u.number,
                            (char *) sp->u.buf->item, sp->u.buf->size);
            break;
        case T_STRING:
            i = write_buffer((sp - 2)->u.buf, (sp - 1)->u.number,
                            sp->u.string, SVALUE_STRLEN(sp));
            break;
        default:
            bad_argument(sp, T_BUFFER | T_STRING | T_NUMBER, 3, F_WRITE_BUFFER);
    }
    free_svalue(sp--, "f_write_buffer");
    free_string_svalue(--sp);
    put_number(i);
}
#endif

#ifdef F_WRITE_FILE
void
f_write_file(void)
{
    int flags = 0;

    if (st_num_arg == 3) {
        flags = (sp--)->u.number;
    }
    flags = write_file((sp - 1)->u.string, sp->u.string, flags);
    free_string_svalue(sp--);
    free_string_svalue(sp);
    put_number(flags);
}
#endif

#ifdef F_DUMP_FILE_DESCRIPTORS
void
f_dump_file_descriptors(void)
{
    outbuffer_t out;

    outbuf_zero(&out);
    dump_file_descriptors(&out);
    outbuf_push(&out);
}
#endif

#ifdef F_MEMORY_INFO
void
f_memory_info(void)
{
    int mem;

    if (st_num_arg == 0) {
	int res, tot;

	if (reserved_area)
	    res = RESERVED_SIZE;
	else
	    res = 0;
	tot = total_prog_block_size +
#ifdef ARRAY_STATS
	    total_array_size +
#endif
	    total_mapping_size +
	    tot_alloc_object_size +
	    tot_alloc_sentence * sizeof(sentence_t) +
	    total_users * sizeof(interactive_t) +
	    show_otable_status(0, -1) +
	    heart_beat_status(0, -1) +
	    add_string_status(0, -1) +
	    print_call_out_usage(0, -1) + res;
	push_number(tot);
	return;
    }
    if (sp->type != T_OBJECT)
	bad_argument(sp, T_OBJECT, 1, F_MEMORY_INFO);
    ob = sp->u.ob;
    if (ob->prog && (ob->prog->ref == 1 || !(ob->flags & O_CLONE)))
	mem = ob->prog->total_size;
    else
	mem = 0;
    mem += (data_size(ob) + sizeof(object_t));
    free_object(ob, "f_memory_info");
    put_number(mem);
}
#endif

#ifdef F_RELOAD_OBJECT
void
f_reload_object(void)
{
    reload_object(sp->u.ob);
    free_object((sp--)->u.ob, "f_reload_object");
}
#endif

#ifdef F_QUERY_SHADOWING
void
f_query_shadowing(void)
{
    if ((sp->type == T_OBJECT) && (ob = sp->u.ob)->shadowing) {
        add_ref(ob->shadowing, "query_shadowing(ob)");
        sp->u.ob = ob->shadowing;
        free_object(ob, "f_query_shadowing");
    } else {
        free_svalue(sp, "f_query_shadowing");
        *sp = const0;
    }
}
#endif

#ifdef F_SET_RESET
void
f_set_reset(void)
{
    if (st_num_arg == 2) {
        (sp - 1)->u.ob->next_reset = current_time + sp->u.number;
        free_object((--sp)->u.ob, "f_set_reset:1");
        sp--;
    } else {
        sp->u.ob->next_reset = current_time + TIME_TO_RESET / 2 +
            rand() % (TIME_TO_RESET / 2);
        free_object((sp--)->u.ob, "f_set_reset:2");
    }
}
#endif

#ifdef F_FLOATP
void
f_floatp(void)
{
    if (sp->type == T_REAL){
        sp->type = T_NUMBER;
        sp->u.number = 1;
    }
    else {
        free_svalue(sp, "f_floatp");
        *sp = const0;
    }
}
#endif

#ifdef F_FLUSH_MESSAGES
void
f_flush_messages(void) {
    if (st_num_arg == 1) {
	if (sp->u.ob->interactive)
	    flush_message(sp->u.ob->interactive);
	pop_stack();
    } else {
	int i;

	for (i = 0; i < max_users; i++) {
	    if (all_users[i] && !(all_users[i]->iflags & CLOSING))
		flush_message(all_users[i]);
	}
    }
}
#endif

#ifdef F_FIRST_INVENTORY
void
f_first_inventory(void)
{
    ob = first_inventory(sp);
    free_svalue(sp, "f_first_inventory");
    if (ob) { put_unrefed_undested_object(ob, "first_inventory"); }
    else *sp = const0;
}
#endif

#ifdef F_NEXT_INVENTORY
void
f_next_inventory(void)
{
    ob = sp->u.ob->next_inv;
    free_object(sp->u.ob, "f_next_inventory");
    while (ob) {
        if (ob->flags & O_HIDDEN) {
	    object_t *old_ob = ob;
            if (object_visible(ob)) {
	        add_ref(old_ob, "next_inventory(ob) : 1");
                sp->u.ob = old_ob;
                return;
	    }
	} else {
	    add_ref(ob, "next_inventory(ob) : 2");
            sp->u.ob = ob;
            return;
	}
        ob = ob->next_inv;
    }
    *sp = const0;
}
#endif

/*
 * This differs from the livings() efun in that this efun only returns
 * objects which have had set_living_name() called as well as
 * enable_commands().  The other major difference is that this is
 * substantially faster.
 */
#ifdef F_NAMED_LIVINGS
void f_named_livings() {
    int i;
    int nob, apply_valid_hide, hide_is_valid = 0;
    object_t *ob, **obtab;
    array_t *vec;

    nob = 0;
    apply_valid_hide = 1;

    obtab = CALLOCATE(max_array_size, object_t *, TAG_TEMPORARY, "named_livings");

    for (i = 0; i < CFG_LIVING_HASH_SIZE; i++) {
	for (ob = hashed_living[i]; ob; ob = ob->next_hashed_living) {
	    if (!(ob->flags & O_ENABLE_COMMANDS))
		continue;
	    if (ob->flags & O_HIDDEN) {
		if (apply_valid_hide) {
		    apply_valid_hide = 0;
		    hide_is_valid = valid_hide(current_object);
		}
		if (hide_is_valid)
		    continue;
	    }
	    if (nob == max_array_size)
		break;
	    obtab[nob++] = ob;
	}
    }

    vec = allocate_empty_array(nob);
    while (--nob >= 0) {
	vec->item[nob].type = T_OBJECT;
	vec->item[nob].u.ob = obtab[nob];
	add_ref(obtab[nob], "livings");
    }

    FREE(obtab);

    push_refed_array(vec);
}
#endif

/* I forgot who wrote this, please claim it :) */
#ifdef F_REMOVE_SHADOW
void
f_remove_shadow(void)
{
    object_t *ob;

    ob = current_object;
    if ( st_num_arg )
	{
	    ob = sp->u.ob;
	    pop_stack();
	}
    if ( ! ob || ! ob->shadowing )
	push_number( 0 );
    else
	{
	    if ( ob->shadowed )
		ob->shadowed->shadowing = ob->shadowing;
	    if ( ob->shadowing )
		ob->shadowing->shadowed = ob->shadowed;
	    ob->shadowing = ob->shadowed = 0;
	    push_number( 1 );
	}
}
#endif

/* This was originally written my Malic for Demon.  I rewrote parts of it
   when I added it (added function support, etc) -Beek */
#ifdef F_QUERY_NOTIFY_FAIL
void
f_query_notify_fail(void) {
    char *p;

    if (command_giver && command_giver->interactive) {
	if (command_giver->interactive->iflags & NOTIFY_FAIL_FUNC) {
	    push_funp(command_giver->interactive->default_err_message.f);
	    return;
	} else if ((p = command_giver->interactive->default_err_message.s)) {
	    sp++;
	    sp->type = T_STRING;
	    sp->subtype = STRING_SHARED;
	    sp->u.string = p;
	    ref_string(p);
	    return;
	}
    }
    push_number(0);
}
#endif

/* Beek again */
#ifdef F_STORE_VARIABLE
void
f_store_variable(void) {
    int idx;
    svalue_t *sv;
    unsigned short type;

    idx = find_global_variable(current_object->prog, (sp-1)->u.string, &type);
    if (idx == -1)
	error("No variable named '%s'!\n", (sp-1)->u.string);
    sv = &current_object->variables[idx];
    free_svalue(sv, "f_store_variable");
    *sv = *sp--;
    free_string_svalue(sp--);
}
#endif

#ifdef F_FETCH_VARIABLE
void
f_fetch_variable(void) {
    int idx;
    svalue_t *sv;
    unsigned short type;

    idx = find_global_variable(current_object->prog, sp->u.string, &type);
    if (idx == -1)
	error("No variable named '%s'!\n", sp->u.string);
    sv = &current_object->variables[idx];
    free_string_svalue(sp--);
    push_svalue(sv);
}
#endif

/* Beek */
#ifdef F_SET_PROMPT
void
f_set_prompt(void) {
    object_t *who;
    if (st_num_arg == 2) {
	who = sp->u.ob;
	pop_stack();
    } else who = command_giver;

    if (!who || who->flags & O_DESTRUCTED || !who->interactive)
	error("Prompts can only be set for interactives.\n");

    /* Future work */
    /* ed() will nuke this; also we have to make sure the string will get
     * freed */
}
#endif

/* Gudu@VR wrote copy_array() and copy_mapping() which this is heavily
 * based on.  I made it into a general copy() efun which incorporates
 * both. -Beek
 */
#ifdef F_COPY
static int depth;

void deep_copy_svalue(svalue_t *, svalue_t *);

array_t *
deep_copy_array(array_t* arg)
{
    array_t *vec;
    int i;

    vec = allocate_empty_array(arg->size);
    for (i = 0; i < arg->size; i++)
	deep_copy_svalue(&arg->item[i], &vec->item[i]);
    return vec;
}

int
doCopy(mapping_t* map, mapping_node_t* elt, mapping_t* dest)
{
    svalue_t *sp;

    sp = find_for_insert(dest, &elt->values[0], 1);
    if (!sp) {
	mapping_too_large();
	return 1;
    }

    deep_copy_svalue(&elt->values[1], sp);
    return 0;
}

mapping_t *
deep_copy_mapping(mapping_t* arg)
{
    mapping_t *map;

    map = allocate_mapping( 0 ); /* this should be fixed.  -Beek */
    mapTraverse( arg, (int (*)()) doCopy, map);
    return map;
}

void
deep_copy_svalue(svalue_t* from, svalue_t* to)
{
    switch (from->type) {
    case T_ARRAY:
    case T_CLASS:
	depth++;
	if (depth > MAX_SAVE_SVALUE_DEPTH) {
	    depth = 0;
	    error("Mappings, arrays and/or classes nested too deep (%d) for copy()\n",
		  MAX_SAVE_SVALUE_DEPTH);
	}
	*to = *from;
	to->u.arr = deep_copy_array( from->u.arr );
	depth--;
	break;
    case T_MAPPING:
	depth++;
	if (depth > MAX_SAVE_SVALUE_DEPTH) {
	    depth = 0;
	    error("Mappings, arrays and/or classes nested too deep (%d) for copy()\n",
		  MAX_SAVE_SVALUE_DEPTH);
	}
	*to = *from;
	to->u.map = deep_copy_mapping( from->u.map );
	depth--;
	break;
    default:
	assign_svalue_no_free( to, from );
    }
}

void f_copy(void)
{
    svalue_t ret;

    depth = 0;
    deep_copy_svalue(sp, &ret);
    free_svalue(sp, "f_copy");
    *sp = ret;
}
#endif

/* Gudu@VR */
/* flag and extra info by Beek */
#ifdef F_FUNCTIONS
void f_functions(void) {
    int i, j, num, index;
    array_t *vec, *subvec;
    runtime_function_u *func_entry;
    compiler_function_t *funp;
    program_t *prog;
    int flag = (sp--)->u.number;
    unsigned short *types;
    char buf[256];
    char *end = EndOf(buf);
    program_t *progp;

    if (sp->u.ob->flags & O_SWAPPED)
	load_ob_from_swap(sp->u.ob);

    progp = sp->u.ob->prog;
    num = progp->num_functions_total;
    if (num && progp->function_table[progp->num_functions_defined-1].name[0]
	== APPLY___INIT_SPECIAL_CHAR)
	num--;

    vec = allocate_empty_array(num);
    i = num;

    while (i--) {
	prog = sp->u.ob->prog;
	index = i;
	func_entry = FIND_FUNC_ENTRY(prog, index);

	/* Walk up the inheritance tree to the real definition */
	while (prog->function_flags[index] & NAME_INHERITED) {
	    prog = prog->inherit[func_entry->inh.offset].prog;
	    index = func_entry->inh.function_index_offset;
	    func_entry = FIND_FUNC_ENTRY(prog, index);
	}

	funp = prog->function_table + func_entry->def.f_index;

	if (flag) {
	    if (prog->type_start && prog->type_start[index] != INDEX_START_NONE)
		types = &prog->argument_types[prog->type_start[index]];
	    else
		types = 0;

	    vec->item[i].type = T_ARRAY;
	    subvec = vec->item[i].u.arr = allocate_empty_array(3 + func_entry->def.num_arg);

	    subvec->item[0].type = T_STRING;
	    subvec->item[0].subtype = STRING_SHARED;
	    subvec->item[0].u.string = ref_string(funp->name);

	    subvec->item[1].type = T_NUMBER;
	    subvec->item[1].subtype = 0;
	    subvec->item[1].u.number = func_entry->def.num_arg;

	    get_type_name(buf, end, funp->type);
	    subvec->item[2].type = T_STRING;
	    subvec->item[2].subtype = STRING_SHARED;
	    subvec->item[2].u.string = make_shared_string(buf);

	    for (j = 0; j < func_entry->def.num_arg; j++) {
		if (types) {
		    get_type_name(buf, end, types[j]);
		    subvec->item[3 + j].type = T_STRING;
		    subvec->item[3 + j].subtype = STRING_SHARED;
		    subvec->item[3 + j].u.string = make_shared_string(buf);
		} else {
		    subvec->item[3 + j].type = T_NUMBER;
		    subvec->item[3 + j].u.number = 0;
		}
	    }
	} else {
	    vec->item[i].type = T_STRING;
	    vec->item[i].subtype = STRING_SHARED;
	    vec->item[i].u.string = ref_string(funp->name);
	}
    }

    pop_stack();
    push_refed_array(vec);
}
#endif

/* Beek */
#ifdef F_VARIABLES
static void
fv_recurse(array_t* arr, int* idx, program_t* prog, int type, int flag)
{
    int i;
    array_t *subarr;
    char buf[256];
    char *end = EndOf(buf);

    for (i = 0; i < prog->num_inherited; i++) {
	fv_recurse(arr, idx, prog->inherit[i].prog,
		   type | prog->inherit[i].type_mod, flag);
    }
    for (i = 0; i < prog->num_variables_defined; i++) {
	if (flag) {
	    arr->item[*idx + i].type = T_ARRAY;
	    subarr = arr->item[*idx + i].u.arr = allocate_empty_array(2);
	    subarr->item[0].type = T_STRING;
	    subarr->item[0].subtype = STRING_SHARED;
	    subarr->item[0].u.string = ref_string(prog->variable_table[i]);
	    get_type_name(buf, end, prog->variable_types[i]);
	    subarr->item[1].type = T_STRING;
	    subarr->item[1].subtype = STRING_SHARED;
	    subarr->item[1].u.string = make_shared_string(buf);
	} else {
	    arr->item[*idx + i].type = T_STRING;
	    arr->item[*idx + i].subtype = STRING_SHARED;
	    arr->item[*idx + i].u.string = ref_string(prog->variable_table[i]);
	}
    }
    *idx += prog->num_variables_defined;
}

void f_variables(void) {
    int idx = 0;
    array_t *arr;
    int flag = (sp--)->u.number;
    program_t *prog = sp->u.ob->prog;

    if (sp->u.ob->flags & O_SWAPPED)
	load_ob_from_swap(sp->u.ob);

    arr = allocate_empty_array(prog->num_variables_total);
    fv_recurse(arr, &idx, prog, 0, flag);

    pop_stack();
    push_refed_array(arr);
}
#endif

/* also Beek */
#ifdef F_HEART_BEATS
void f_heart_beats(void) {
    push_refed_array(get_heart_beats());
}
#endif

/*Aleas@Nightmare */
#ifdef F_TERMINAL_COLOUR
/* A fast implementation of the Nightmare color support.

   [Ed note: These codes were actually used on Discworld
   before Nightmare]

   Rewritten several times, since Beek wants it to be
   perfect :)

   Takes a string and a mapping as args. The string is
   exploded using "%^" as delimiter, then all keys of
   the mapping found in the resulting array are replaced
   by their values. Afterwards a string imploded from
   the array is returned.

   No actual string copying is done except for the
   creation of the final string and a temporary copy of
   the input string to avoid destruction of shared
   input strings. An array of pointers to the segments
   of the string is compared against the mapping keys
   and replaced with a pointer to the value belonging to
   that key where matches are found.

   After the replacement pass the result string is created
   from the pointer array.

   Further speed is gained by the fact that no parsing is
   done if the input string does not contain any "%^"
   delimiter sequence.

   by Aleas@Nightmare, dec-94 */

/* number of input string segments, if more, it still works, but a
   _slow_ realloc is required */
#define NSTRSEGS 32
#define TC_FIRST_CHAR '%'
#define TC_SECOND_CHAR '^'

int at_end(int i, int imax, int z, int *lens) {
    if (z + 1 != lens[i])
	return 0;
    for (i++; i < imax; i++) {
	if (lens[i] > 0)
	    return 0;
    }
    return 1;
}

void
f_terminal_colour(int num_arg, int instruction)
{
    char *instr, *cp, *savestr = NULL, *deststr, **parts;
    int num, i, j, k, col, space, *lens, maybe_at_end;
    int space_garbage = 0;
    mapping_node_t *elt, **mtab;
    int tmp;
    int wrap = 0;
    int indent = 0;

    if (st_num_arg >= 3) {
	if (st_num_arg == 4)
	    indent = (sp--)->u.number;
	wrap = (sp--)->u.number;
	if (wrap < 2 && wrap != 0) wrap = 2;
	if (indent < 0 || indent >= wrap - 1)
	    indent = wrap - 2;
    }

    cp = instr = (sp-1)->u.string;
    do {
	cp = strchr(cp, TC_FIRST_CHAR);
	if (cp)
	{
	    if (cp[1] == TC_SECOND_CHAR)
	    {
		savestr = string_copy(instr, "f_terminal_colour");
		cp = savestr + ( cp - instr );
		instr = savestr;
		break;
	    }
	    cp++;
	}
    } while (cp);
    if (cp == NULL) {
	if (wrap) {
	    num = 1;
	    parts = CALLOCATE(1, char *, TAG_TEMPORARY, "f_terminal_colour: parts");
	    parts[0] = instr;
	    savestr = 0;
	} else {
	    pop_stack(); /* no delimiter in string, so return the original */
	    return;
	}
    } else {
        /* here we have something to parse */

	parts = CALLOCATE(NSTRSEGS, char *, TAG_TEMPORARY, "f_terminal_colour: parts");
	if (cp - instr) {	/* starting seg, if not delimiter */
	    num = 1;
	    parts[0] = instr;
	    *cp = 0;
	} else
	    num = 0;
	while (cp) {
	    cp += 2;
	    instr = cp;
	    do {
		cp = strchr(cp,TC_FIRST_CHAR);
		if (cp) {
		    if (cp[1] == TC_SECOND_CHAR)
			break;
		    cp++;
		}
	    } while (cp);
	    if (cp) {
		*cp = 0;
		if (cp > instr) {
		    parts[num] = instr;
		    num++;
		    if (num % NSTRSEGS == 0)
			parts = RESIZE(parts, num + NSTRSEGS, char *,
				       TAG_TEMPORARY, "f_terminal_colour: parts realloc");
		}
	    }
	}
	if (*instr)	/* trailing seg, if not delimiter */
	    parts[num++] = instr;
    }

    /* Could keep track of the lens as we create parts, removing the need
       for a strlen() below */
    lens = CALLOCATE(num, int, TAG_TEMPORARY, "f_terminal_colour: lens");
    mtab = sp->u.map->table;

    /* Do the the pointer replacement and calculate the lengths */
    col = 0;
    space = 0;
    maybe_at_end = 0;
    for (j = i = 0, k = sp->u.map->table_size; i < num; i++)
    {
	int len = 0;

	if ((cp = findstring(parts[i]))) {
	    tmp = MAP_POINTER_HASH(cp);
	    for (elt = mtab[tmp & k]; elt; elt = elt->next)
		if ( elt->values->type == T_STRING &&
		     (elt->values + 1)->type == T_STRING &&
		     cp == elt->values->u.string) {
		    parts[i] = (elt->values + 1)->u.string;
		    /* Negative indicates don't count for wrapping */
		    len = SVALUE_STRLEN(elt->values + 1);
		    if (wrap) len = -len;
		    break;
		}
	    if (!elt)
		len = SHARED_STRLEN(cp);
	} else {
	    len = strlen(parts[i]);
	}
	lens[i] = len;
	if (len > 0) {
	    if (maybe_at_end) {
		if (j + indent > max_string_length) {
		    /* this string no longer counts, so we are still in
		       a maybe_at_end condition.  This means we will end
		       up truncating the rest of the fragments too, since
		       the indent will never fit. */
		    lens[i] = 0;
		    len = 0;
		} else {
		    j += indent;
		    col += indent;
		    maybe_at_end = 0;
		}
	    }
	    j += len;
	    if (j > max_string_length) {
		lens[i] -= j - max_string_length;
		j = max_string_length;
	    }
	    if (wrap) {
		int z;
		char *p = parts[i];
		for (z = 0; z < lens[i]; z++) {
		    char c = p[z];
		    if (c == '\n') {
			col = 0;
		    } else {
			col++;
			if (c == ' ')
			    space = col;
			if (col == wrap+1) {
			    if (space) {
				col -= space;
				space = 0;
			    } else {
				j++;
				col = 1;
			    }
			} else
			    continue;
		    }
		    /* If we get here, we ended a line */
		    if (z + 1 != lens[i]) {
			j += indent;
			col += indent;
		    } else
			maybe_at_end = 1;

		    if (j > max_string_length) {
			lens[i] -= (j - max_string_length);
			j = max_string_length;
			if (lens[i] < z) {
			    /* must have been ok
			       or we wouldn't be here */
			    lens[i] = z;
			    break;
			}
		    }
		}
	    }
	} else {
	    j += -len;
	    if (j > max_string_length) {
		lens[i] = -(-(lens[i]) - (j - max_string_length));
		j = max_string_length;
	    }
	}
    }

    /* now we have the final string in parts and length in j.
       let's compose it, wrapping if necessary */
    cp = deststr = new_string(j, "f_terminal_colour: deststr");
    if (wrap) {
	/* FIXME */
	char *tmp = new_string(8192, "f_terminal_colour: wrap");
	char *pt = tmp;

	col = 0;
	space = 0;
	for (i = 0; i < num; i++) {
	    int kind;
	    int len;
	    int l = lens[i];
	    char *p = parts[i];
	    if (l < 0) {
		memcpy(pt, p, -l);
		pt += -l;
		space_garbage += -l; /* Number of chars due to ignored junk
					since last space */
		continue;
	    }
	    for (k = 0; k < lens[i]; k++) {
		int n;
		char c = p[k];
		*pt++ = c;
		if (c == '\n') {
		    col = 0;
		    kind = 0;
		} else {
		    col++;
		    if (c == ' ') {
			space = col;
			space_garbage = 0;
		    }
		    if (col == wrap+1) {
			if (space) {
			    col -= space;
			    space = 0;
			    kind = 1;
			} else {
			    col = 1;
			    kind = 2;
			}
		    } else
			continue;
		}
		/* If we get here, we ended a line */
		len = (kind == 1 ? col + space_garbage : col);
		n = (pt - tmp) - len;
		memcpy(cp, tmp, n);
		cp += n;
		if (kind == 1) {
		    /* replace the space */
		    cp[-1] = '\n';
		}
		if (kind == 2) {
		    /* need to insert a newline */
		    *cp++ = '\n';
		}
		memmove(tmp, tmp + n, len);
		pt = tmp + len;
		if (!at_end(i, num, k, lens)) {
		    memset(cp, ' ', indent);
		    cp += indent;
		    col += indent;
		}
	    }
	}
	memcpy(cp, tmp, pt - tmp);
	cp += pt - tmp;
	FREE_MSTR(tmp);
    } else {
	for (i = 0; i < num; i++) {
	    memcpy(cp, parts[i], lens[i]);
	    cp += lens[i];
	}
    }
    *cp = 0;
    FREE(lens);
    FREE(parts);
    if (savestr)
	FREE_MSTR(savestr);
    /* now we have what we want */
    pop_stack();

    free_string_svalue(sp);
    sp->type = T_STRING;
    sp->subtype = STRING_MALLOC;
    sp->u.string = deststr;
}
#endif

#ifdef F_FILE_LENGTH
/*
 * file_length() efun, returns the number of lines in a file.
 * Returns -1 if no privs or file doesn't exist.
 */
int
file_length(char* file)
{
  struct stat st;
  FILE *f;
  int ret = 0;
  int num;
  char buf[2049];
  char *p, *newp;

  file = check_valid_path(file, current_object, "file_size", 0);

  if (!file) return -1;
  if (stat(file, &st) == -1)
      return -1;
  if (st.st_mode & S_IFDIR)
      return -2;
  if (!(f = fopen(file, "r")))
      return -1;

  do {
      num = fread(buf, 1, 2048, f);
      p = buf - 1;
      while ((newp = memchr(p + 1, '\n', num))) {
	  num -= (newp - p);
	  p = newp;
	  ret++;
      }
  } while (!feof(f));

  fclose(f);
  return ret;
} /* end of file_length() */

void
f_file_length(void)
{
    int l;

    l = file_length(sp->u.string);
    pop_stack();
    push_number(l);
}
#endif

#ifdef F_UPPER_CASE
void
f_upper_case(void)
{
    register char *str;

    str = sp->u.string;
    /* find first upper case letter, if any */
    for (; *str; str++) {
	if (islower(*str)) {
	    int l = str - sp->u.string;
	    unlink_string_svalue(sp);
	    str = sp->u.string + l;
	    *str -= 'a' - 'A';
	    for (str++; *str; str++) {
		if (islower(*str))
		    *str -= 'a' - 'A';
	    }
	    return;
	}
    }
}
#endif

#ifdef F_REPLACEABLE
void f_replaceable(void) {
    program_t *prog;
    int i, j, num, numignore;
    char **ignore;

    if (st_num_arg == 2) {
	numignore = sp->u.arr->size;
	if (numignore)
	    ignore = CALLOCATE(numignore, char *, TAG_TEMPORARY, "replaceable");
	else
	    ignore = 0;
	for (i = 0; i < numignore; i++) {
	    if (sp->u.arr->item[i].type == T_STRING)
		ignore[i] = findstring(sp->u.arr->item[i].u.string);
	    else
		ignore[i] = 0;
	}
	prog = (sp-1)->u.ob->prog;
    } else {
	numignore = 1;
	ignore = CALLOCATE(1, char *, TAG_TEMPORARY, "replaceable");
	ignore[0] = findstring(APPLY_CREATE);
	prog = sp->u.ob->prog;
    }

    num = prog->num_functions_total;

    for (i = 0; i < num; i++) {
	if (prog->function_flags[i] & (NAME_INHERITED | NAME_NO_CODE)) continue;
	for (j = 0; j < numignore; j++)
	    if (ignore[j] == prog->function_table[FIND_FUNC_ENTRY(prog, i)->def.f_index].name)
		break;
	if (j == numignore)
	    break;
    }
    if (st_num_arg == 2)
	free_array((sp--)->u.arr);
    FREE(ignore);
    free_svalue(sp, "f_replaceable");
    put_number(i == num);
}
#endif

#ifdef F_PROGRAM_INFO
void f_program_info(void) {
    int func_size = 0;
    int string_size = 0;
    int var_size = 0;
    int inherit_size = 0;
    int prog_size = 0;
    int hdr_size = 0;
    int class_size = 0;
    int type_size = 0;
    int total_size = 0;
    object_t *ob;
    program_t *prog;
    outbuffer_t out;
    int i, n;

    if (st_num_arg == 1) {
	ob = sp->u.ob;
	prog = ob->prog;
	if (!(ob->flags & (O_CLONE|O_SWAPPED))) {
	    hdr_size += sizeof(program_t);
	    prog_size += prog->program_size;
	    func_size += 2 * prog->num_functions_total; /* function flags */
#ifdef COMPRESS_FUNCTION_TABLES
	    /* compressed table header */
	    func_size += sizeof(compressed_offset_table_t) - 1;
	    /* it's entries */
	    func_size += (prog->function_compressed->first_defined - prog->function_compressed->num_compressed);
	    /* offset table */
	    func_size += sizeof(runtime_function_u) * (prog->num_functions_total - prog->function_compressed->num_deleted);
#else
	    /* offset table */
	    func_size += prog->num_functions_total * sizeof(runtime_function_u);
#endif
	    /* definitions */
	    func_size += prog->num_functions_defined * sizeof(compiler_function_t);
	    string_size += prog->num_strings * sizeof(char *);
	    var_size += prog->num_variables_defined * (sizeof(char *) + sizeof(unsigned short));
	    inherit_size += prog->num_inherited * sizeof(inherit_t);
	    if (prog->num_classes)
		class_size += prog->num_classes * sizeof(class_def_t) + (prog->classes[prog->num_classes - 1].index + prog->classes[prog->num_classes - 1].size) * sizeof(class_member_entry_t);
	    type_size += prog->num_functions_defined * sizeof(short);
	    n = 0;
	    for (i = 0; i < prog->num_functions_defined; i++) {
		int start;
		unsigned short *ts = prog->type_start;
		int ri;

		if (!ts) continue;
		start = ts[i];
		if (start == INDEX_START_NONE)
		    continue;
		ri = prog->function_table[i].runtime_index;
		start += FIND_FUNC_ENTRY(prog, ri)->def.num_arg;
		if (start > n)
		    n = start;
	    }
	    type_size += n * sizeof(short);
	    total_size += prog->total_size;
	}
	pop_stack();
    } else {
	for (ob = obj_list; ob; ob = ob->next_all) {
	    if (ob->flags & (O_CLONE|O_SWAPPED)) continue;
	    prog = ob->prog;
	    hdr_size += sizeof(program_t);
	    prog_size += prog->program_size;
	    func_size += prog->num_functions_total; /* function flags */
#ifdef COMPRESS_FUNCTION_TABLES
	    /* compressed table header */
	    func_size += sizeof(compressed_offset_table_t) - 1;
	    /* it's entries */
	    func_size += (prog->function_compressed->first_defined - prog->function_compressed->num_compressed);
	    /* offset table */
	    func_size += sizeof(runtime_function_u) * (prog->num_functions_total - prog->function_compressed->num_deleted);
#else
	    /* offset table */
	    func_size += prog->num_functions_total * sizeof(runtime_function_u);
#endif
	    /* definitions */
	    func_size += prog->num_functions_defined * sizeof(compiler_function_t);
	    string_size += prog->num_strings * sizeof(char *);
	    var_size += prog->num_variables_defined * (sizeof(char *) + sizeof(unsigned short));
	    inherit_size += prog->num_inherited * sizeof(inherit_t);
	    if (prog->num_classes)
		class_size += prog->num_classes * sizeof(class_def_t) + (prog->classes[prog->num_classes - 1].index + prog->classes[prog->num_classes - 1].size) * sizeof(class_member_entry_t);
	    type_size += prog->num_functions_defined * sizeof(short);
	    n = 0;
	    for (i = 0; i < prog->num_functions_defined; i++) {
		int start;
		int ri;

		unsigned short *ts = prog->type_start;
		if (!ts) continue;
		start = ts[i];
		if (start == INDEX_START_NONE)
		    continue;
		ri = prog->function_table[i].runtime_index;
		start += FIND_FUNC_ENTRY(prog, ri)->def.num_arg;
		if (start > n)
		    n = start;
	    }
	    type_size += n * sizeof(short);
	    total_size += prog->total_size;
	}
    }

    outbuf_zero(&out);

    outbuf_addv(&out, "\nheader size: %i\n", hdr_size);
    outbuf_addv(&out, "code size: %i\n", prog_size);
    outbuf_addv(&out, "function size: %i\n", func_size);
    outbuf_addv(&out, "string size: %i\n", string_size);
    outbuf_addv(&out, "var size: %i\n", var_size);
    outbuf_addv(&out, "class size: %i\n", class_size);
    outbuf_addv(&out, "inherit size: %i\n", inherit_size);
    outbuf_addv(&out, "saved type size: %i\n\n", type_size);

    outbuf_addv(&out, "total size: %i\n", total_size);

    outbuf_push(&out);
}
#endif

/* Magician - 08May95
 * int remove_interactive(object ob)
 * If the object isn't destructed and is interactive, then remove it's
 * interactivity and disconnect it.  (useful for exec()ing to an already
 * interactive object, ie, Linkdead reconnection)
 */

#ifdef F_REMOVE_INTERACTIVE
void f_remove_interactive(void) {
    if( (sp->u.ob->flags & O_DESTRUCTED) || !(sp->u.ob->interactive) ) {
	free_object(sp->u.ob, "f_remove_interactive");
	*sp = const0;
    } else {
        remove_interactive(sp->u.ob, 0);
	/* It may have been dested */
	if (sp->type == T_OBJECT)
	    free_object(sp->u.ob, "f_remove_interactive");
	*sp = const1;
    }
}
#endif

/* Zakk - August 23 1995
 * return the port number the interactive object used to connect to the
 * mud.
 */
#ifdef F_QUERY_IP_PORT
int
query_ip_port(object_t* ob)
{
    if (!ob || ob->interactive == 0)
	return 0;
    return ob->interactive->local_port;
}

void
f_query_ip_port(void)
{
    int tmp;

    if (st_num_arg) {
	tmp = query_ip_port(sp->u.ob);
	free_object(sp->u.ob, "f_query_ip_port");
    } else {
	tmp = query_ip_port(command_giver);
	sp++;
    }
    put_number(tmp);
}
#endif

/*
** John Viega (rust@lima.imaginary.com) Jan, 1996
** efuns for doing time zone conversions.  Much friendlier
** than doing all the lookup tables in LPC.
** most muds have traditionally just used an offset of the
** mud time or GMT, and this isn't always correct.
*/

#ifdef F_ZONETIME

char *
set_timezone (char * timezone)
{
  char put_tz[20];
  char *old_tz;

  old_tz = getenv("TZ");
  sprintf (put_tz, "TZ=%s", timezone);
  putenv (put_tz);
  tzset ();
  return old_tz;
}

void
reset_timezone (char *old_tz)
{
  int  i = 0;
  int  env_size = 0;
  char put_tz[20];

  if (!old_tz)
    {
      while (environ[env_size] != NULL)
        {
          if (strlen (environ[env_size]) > 3 && environ[env_size][2] == '='
             && environ[env_size][1] == 'Z' && environ[env_size][0] == 'T')
            {
              i = env_size;
            }
          env_size++;
        }
      if ((i+1) == env_size)
        {
          environ[i] = NULL;
        }
      else
        {
          environ[i] = environ[env_size-1];
          environ[env_size-1] = NULL;
        }
    }
  else
    {
      sprintf (put_tz, "TZ=%s", old_tz);
      putenv (put_tz);
    }
  tzset ();
}

void
f_zonetime(void)
{
  char *timezone, *old_tz;
  char *retv;
  int  time_val;
  int  len;

  time_val   = sp->u.number;
  pop_stack ();
  timezone   = sp->u.string;
  pop_stack ();

  old_tz = set_timezone (timezone);
  retv = ctime ((time_t *)&time_val);
  len  = strlen (retv);
  retv[len-1] = '\0';
  reset_timezone (old_tz);
  push_malloced_string (string_copy(retv, "zonetime"));

}
#endif

#ifdef F_IS_DAYLIGHT_SAVINGS_TIME
void
f_is_daylight_savings_time(void)
{
  struct tm *t;
  int       time_to_check;
  char      *timezone;
  char      *old_tz;

  time_to_check = sp->u.number;
  pop_stack ();
  timezone = sp->u.string;
  pop_stack ();

  old_tz = set_timezone (timezone);

  t = localtime ((time_t *)&time_to_check);

  push_number ((t->tm_isdst) > 0);

  reset_timezone (old_tz);
}
#endif

#ifdef F_DEBUG_MESSAGE
void f_debug_message(void) {
    debug_message("%s\n", sp->u.string);
    free_string_svalue(sp--);
}
#endif

#ifdef F_FUNCTION_OWNER
void f_function_owner(void) {
    object_t *owner = sp->u.fp->hdr.owner;

    free_funp(sp->u.fp);
    put_unrefed_object(owner, "f_function_owner");
}
#endif

#ifdef F_REPEAT_STRING
void f_repeat_string(void) {
    char *str;
    int repeat, len;
    char *ret, *p;
    int i;

    repeat = (sp--)->u.number;
    if (repeat <= 0) {
	free_string_svalue(sp);
	sp->type = T_STRING;
	sp->subtype = STRING_CONSTANT;
	sp->u.string = "";
    } else if (repeat != 1) {
	str = sp->u.string;
	len = SVALUE_STRLEN(sp);
	p = ret = new_string(len * repeat, "f_repeat_string");
	for (i = 0; i < repeat; i++) {
	    memcpy(p, str, len);
	    p += len;
	}
	*p = 0;
	free_string_svalue(sp);
	sp->type = T_STRING;
	sp->subtype = STRING_MALLOC;
	sp->u.string = ret;
    }
}
#endif

#ifdef F_MEMORY_SUMMARY
static int memory_share(svalue_t *);

static int
node_share(mapping_t* m, mapping_node_t* elt, void* tp)
{
    int *t = (int *)tp;

    *t += sizeof(mapping_node_t) - 2*sizeof(svalue_t);
    *t += memory_share(&elt->values[0]);
    *t += memory_share(&elt->values[1]);

    return 0;
}

static int
memory_share(svalue_t* sv)
{
    int i, total = sizeof(svalue_t);
    int subtotal;

    switch (sv->type) {
    case T_STRING:
	switch (sv->subtype) {
	case STRING_MALLOC:
	    return total +
		(1 + COUNTED_STRLEN(sv->u.string) + sizeof(malloc_block_t))/
		(COUNTED_REF(sv->u.string));
	case STRING_SHARED:
	    return total +
		(1 + COUNTED_STRLEN(sv->u.string) + sizeof(block_t))/
		(COUNTED_REF(sv->u.string));
	}
	break;
    case T_ARRAY:
    case T_CLASS:
	/* first svalue is stored inside the array struct, so sizeof(array_t)
	 * includes one svalue.
	 */
	subtotal = sizeof(array_t) - sizeof(svalue_t);
	for (i = 0; i < sv->u.arr->size; i++)
	    subtotal += memory_share(&sv->u.arr->item[i]);
	return total + subtotal/sv->u.arr->ref;
    case T_MAPPING:
	subtotal = sizeof(mapping_t);
	mapTraverse(sv->u.map, node_share, &subtotal);
	return total + subtotal/sv->u.map->ref;
    case T_FUNCTION:
    {
	svalue_t tmp;
	tmp.type = T_ARRAY;
	tmp.u.arr = sv->u.fp->hdr.args;

	if (tmp.u.arr)
	    subtotal = sizeof(funptr_hdr_t) + memory_share(&tmp) - sizeof(svalue_t);
	else
	    subtotal = sizeof(funptr_hdr_t);
	switch (sv->u.fp->hdr.type) {
	case FP_EFUN:
	    subtotal += sizeof(efun_ptr_t);
	    break;
	case FP_LOCAL | FP_NOT_BINDABLE:
	    subtotal += sizeof(local_ptr_t);
	    break;
	case FP_SIMUL:
	    subtotal += sizeof(simul_ptr_t);
	    break;
	case FP_FUNCTIONAL:
	case FP_FUNCTIONAL | FP_NOT_BINDABLE:
	    subtotal += sizeof(functional_t);
	    break;
	}
	return total + subtotal/sv->u.fp->hdr.ref;
    }
    case T_BUFFER:
	/* first byte is stored inside the buffer struct */
	return total + (sizeof(buffer_t) + sv->u.buf->size - 1)/sv->u.buf->ref;
    }
    return total;
}


/*
 * The returned mapping is:
 *
 * map["program name"]["variable name"] = memory usage
 */
static void
fms_recurse(mapping_t* map, object_t* ob, int* idx, program_t* prog)
{
    int i;
    svalue_t *entry;
    svalue_t sv;

    sv.type = T_STRING;
    sv.subtype = STRING_SHARED;

    for (i = 0; i < prog->num_inherited; i++)
	fms_recurse(map, ob, idx, prog->inherit[i].prog);

    for (i = 0; i < prog->num_variables_defined; i++) {
	int size = memory_share(ob->variables + *idx + i);

	sv.u.string = prog->variable_table[i];
	entry = find_for_insert(map, &sv, 0);
	entry->u.number += size;
    }
    *idx += prog->num_variables_defined;
}

void f_memory_summary(void) {
    mapping_t *result = allocate_mapping(8);
    object_t *ob;
    int idx;
    svalue_t sv;

    sv.type = T_STRING;
    sv.subtype = STRING_SHARED;

    for (ob = obj_list; ob; ob = ob->next_all) {
	svalue_t *entry;

	if (ob->flags & O_SWAPPED)
	    load_ob_from_swap(ob);

	sv.u.string = ob->prog->name;
	entry = find_for_insert(result, &sv, 0);
	if (entry->type == T_NUMBER) {
	    entry->type = T_MAPPING;
	    entry->u.map = allocate_mapping(8);
	}
	idx = 0;
	fms_recurse(entry->u.map, ob, &idx, ob->prog);
    }
    push_refed_mapping(result);
}
#endif	/* F_MEMORY_SUMMARY */

#ifdef F_DEBUG_INFO
void
f_debug_info(void)
{
    svalue_t *arg;
    outbuffer_t out;
    object_t *ob;

    outbuf_zero(&out);
    arg = sp - 1;
    switch (arg[0].u.number) {
    case 0:
	{
	    int i, flags;
	    object_t *obj2;

	    ob = arg[1].u.ob;
	    flags = ob->flags;
	   outbuf_addv(&out, "O_HEART_BEAT      : %s\n",
			flags & O_HEART_BEAT ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_IS_WIZARD       : %s\n",
			flags & O_IS_WIZARD ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_ENABLE_COMMANDS : %s\n",
			flags & O_ENABLE_COMMANDS ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_CLONE           : %s\n",
			flags & O_CLONE ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_VIRTUAL         : %s\n",
			flags & O_VIRTUAL ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_DESTRUCTED      : %s\n",
			flags & O_DESTRUCTED ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_SWAPPED         : %s\n",
			flags & O_SWAPPED ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_ONCE_INTERACTIVE: %s\n",
			flags & O_ONCE_INTERACTIVE ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_RESET_STATE     : %s\n",
			flags & O_RESET_STATE ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_WILL_CLEAN_UP   : %s\n",
			flags & O_WILL_CLEAN_UP ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "O_WILL_RESET: %s\n",
			flags & O_WILL_RESET ? "TRUE" : "FALSE");
	   outbuf_addv(&out, "next_reset  : %d\n", ob->next_reset);
	   outbuf_addv(&out, "time_of_ref : %d\n", ob->time_of_ref);
	   outbuf_addv(&out, "ref         : %d\n", ob->ref);
	   outbuf_addv(&out, "swap_num    : %d\n", ob->swap_num);
	   outbuf_addv(&out, "name        : '/%s'\n", ob->name);
	   outbuf_addv(&out, "next_all    : OBJ(/%s)\n",
			ob->next_all ? ob->next_all->name : "NULL");
	    if (obj_list == ob)
		outbuf_add(&out, "This object is the head of the object list.\n");
	    for (obj2 = obj_list, i = 1; obj2; obj2 = obj2->next_all, i++)
		if (obj2->next_all == ob) {
		   outbuf_addv(&out, "Previous object in object list: OBJ(/%s)\n",
				obj2->name);
		   outbuf_addv(&out, "position in object list:%d\n", i);
		}
	    break;
	}
    case 1:
	ob = arg[1].u.ob;
	if (ob->flags & O_SWAPPED) {
	    outbuf_add(&out, "Swapped\n");
	    break;
	}
	outbuf_addv(&out, "program ref's %d\n", ob->prog->ref);
	outbuf_addv(&out, "Name /%s\n", ob->prog->name);
	outbuf_addv(&out, "program size %d\n",
		    ob->prog->program_size);
	outbuf_addv(&out, "runtime function table %d (%d) \n",
		    ob->prog->num_functions_total,
		    ob->prog->num_functions_total * (sizeof(runtime_function_u)+1));
	outbuf_addv(&out, "compiler function table %d (%d) \n",
		    ob->prog->num_functions_defined,
		    ob->prog->num_functions_defined * sizeof(compiler_function_t));
	outbuf_addv(&out, "num strings %d\n", ob->prog->num_strings);
	outbuf_addv(&out, "num vars %d (%d)\n", ob->prog->num_variables_defined,
		    ob->prog->num_variables_defined * (sizeof(char *) + sizeof(short)));
	outbuf_addv(&out, "num inherits %d (%d)\n", ob->prog->num_inherited,
		    ob->prog->num_inherited * sizeof(inherit_t));
	outbuf_addv(&out, "total size %d\n", ob->prog->total_size);
	break;
    case 2:
	{
	    int i;
	    ob = arg[1].u.ob;
	    for (i=0; i<ob->prog->num_variables_total; i++) {
		/* inefficient, but: */
		outbuf_addv(&out, "%s: ", variable_name(ob->prog, i));
		svalue_to_string(&ob->variables[i], &out, 2, 0, 0);
		outbuf_add(&out, "\n");
	    }
	    break;
	}
    default:
	bad_arg(1, F_DEBUG_INFO);
    }
    pop_stack();
    pop_stack();
    outbuf_push(&out);
}
#endif

#ifdef F_REFS
void
f_refs(void)
{
    int r;

    switch (sp->type) {
    case T_MAPPING:
	r = sp->u.map->ref;
	break;
    case T_CLASS:
    case T_ARRAY:
	r = sp->u.arr->ref;
	break;
    case T_OBJECT:
	r = sp->u.ob->ref;
	break;
    case T_FUNCTION:
	r = sp->u.fp->hdr.ref;
	break;
    case T_BUFFER:
	r = sp->u.buf->ref;
	break;
    default:
	r = 0;
	break;
    }
    free_svalue(sp, "f_refs");
    put_number(r - 1);		/* minus 1 to compensate for being arg of
				 * refs() */
}
#endif

#if (defined(DEBUGMALLOC) && defined(DEBUGMALLOC_EXTENSIONS))
#ifdef F_DEBUGMALLOC
void
f_debugmalloc(void)
{
    char *res;

    res = dump_debugmalloc((sp - 1)->u.string, sp->u.number);
    free_string_svalue(--sp);
    sp->subtype = STRING_MALLOC;
    sp->u.string = res;
}
#endif

#ifdef F_SET_MALLOC_MASK
void
f_set_malloc_mask(void)
{
    set_malloc_mask((sp--)->u.number);
}
#endif

#ifdef F_CHECK_MEMORY
void
f_check_memory(void)
{
    check_all_blocks((sp--)->u.number);
}
#endif
#endif				/* (defined(DEBUGMALLOC) &&
				 * defined(DEBUGMALLOC_EXTENSIONS)) */

#ifdef F_TRACE
void
f_trace(void)
{
    int ot = -1;

    if (command_giver && command_giver->interactive) {
        ot = command_giver->interactive->trace_level;
        command_giver->interactive->trace_level = sp->u.number;
    }
    sp->u.number = ot;
}
#endif

#ifdef F_TRACEPREFIX
void
f_traceprefix(void)
{
    char *old = 0;

    if (command_giver && command_giver->interactive) {
        old = command_giver->interactive->trace_prefix;
        if (sp->type & T_STRING) {
            command_giver->interactive->trace_prefix =
                make_shared_string(sp->u.string);
            free_string_svalue(sp);
	} else
            command_giver->interactive->trace_prefix = 0;
    }
    if (old) {
        put_shared_string(old);
    } else *sp = const0;
}
#endif

#ifdef F_CRYPT
#define SALT_LEN	8

void
f_crypt(void)
{
    char *res, *p, salt[SALT_LEN + 1];
    char *choice =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ./";

    if (sp->type == T_STRING && SVALUE_STRLEN(sp) >= 2) {
	p = sp->u.string;
    } else {
	int i;

	for (i = 0; i < SALT_LEN; i++)
	    salt[i] = choice[rand() % strlen(choice)];

	salt[SALT_LEN] = 0;
	p = salt;
    }

    res = string_copy(crypt((sp-1)->u.string, p), "f_crypt");
    pop_stack();
    free_string_svalue(sp);
    sp->subtype = STRING_MALLOC;
    sp->u.string = res;
}
#endif

#ifdef F_OLDCRYPT
void
f_oldcrypt(void) {
    char *res, salt[3];
    char *choice =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ./";

    if (sp->type == T_STRING && SVALUE_STRLEN(sp) >= 2) {
        salt[0] = sp->u.string[0];
        salt[1] = sp->u.string[1];
        free_string_svalue(sp--);
    } else {
        salt[0] = choice[rand() % strlen(choice)];
        salt[1] = choice[rand() % strlen(choice)];
        pop_stack();
    }
    salt[2] = 0;
    res = string_copy(crypt(sp->u.string, salt), "f_crypt");
    free_string_svalue(sp);
    sp->subtype = STRING_MALLOC;
    sp->u.string = res;
}
#endif

#ifdef F_LOCALTIME
/* FIXME: most of the #ifdefs here should be based on configure checks
   instead.  Same for rusage() */
void
f_localtime(void)
{
    struct tm *tm;
    array_t *vec;
    time_t lt;

#ifdef sequent
    struct timezone tz;
#endif

    lt = sp->u.number;
    tm = localtime(&lt);

    vec = allocate_empty_array(10);
    vec->item[LT_SEC].type = T_NUMBER;
    vec->item[LT_SEC].u.number = tm->tm_sec;
    vec->item[LT_MIN].type = T_NUMBER;
    vec->item[LT_MIN].u.number = tm->tm_min;
    vec->item[LT_HOUR].type = T_NUMBER;
    vec->item[LT_HOUR].u.number = tm->tm_hour;
    vec->item[LT_MDAY].type = T_NUMBER;
    vec->item[LT_MDAY].u.number = tm->tm_mday;
    vec->item[LT_MON].type = T_NUMBER;
    vec->item[LT_MON].u.number = tm->tm_mon;
    vec->item[LT_YEAR].type = T_NUMBER;
    vec->item[LT_YEAR].u.number = tm->tm_year + 1900;
    vec->item[LT_WDAY].type = T_NUMBER;
    vec->item[LT_WDAY].u.number = tm->tm_wday;
    vec->item[LT_YDAY].type = T_NUMBER;
    vec->item[LT_YDAY].u.number = tm->tm_yday;
    vec->item[LT_GMTOFF].type = T_NUMBER;
    vec->item[LT_ZONE].type = T_STRING;
    vec->item[LT_ZONE].subtype = STRING_MALLOC;
#if defined(BSD42) || defined(apollo) || defined(_AUX_SOURCE) \
	|| defined(OLD_ULTRIX) || defined(__CYGWIN__)
    /* 4.2 BSD doesn't seem to provide any way to get these last two values */
    vec->item[LT_GMTOFF].u.number = 0;
    vec->item[LT_ZONE].type = T_NUMBER;
    vec->item[LT_ZONE].u.number = 0;
#else				/* BSD42 */
#if defined(sequent)
    vec->item[LT_GMTOFF].u.number = 0;
    gettimeofday(NULL, &tz);
    vec->item[LT_GMTOFF].u.number = tz.tz_minuteswest;
    vec->item[LT_ZONE].u.string =
	string_copy(timezone(tz.tz_minuteswest, tm->tm_isdst), "f_localtime");
#else				/* sequent */
#if (defined(hpux) || defined(_SEQUENT_) || defined(_AIX) || defined(SunOS_5) \
	|| defined(SVR4) || defined(sgi) || defined(linux) || defined(cray) \
	|| defined(LATTICE) || defined(SCO))
    if (!tm->tm_isdst) {
	vec->item[LT_GMTOFF].u.number = timezone;
	vec->item[LT_ZONE].u.string = string_copy(tzname[0], "f_localtime");
    } else {
#if (defined(_AIX) || defined(hpux) || defined(linux) || defined(cray) \
	|| defined(LATTICE))
	vec->item[LT_GMTOFF].u.number = timezone;
#else
	vec->item[LT_GMTOFF].u.number = altzone;
#endif
	vec->item[LT_ZONE].u.string = string_copy(tzname[1], "f_localtime");
    }
#else
#ifndef WIN32
    vec->item[LT_GMTOFF].u.number = tm->tm_gmtoff;
    vec->item[LT_ZONE].u.string = string_copy(tm->tm_zone, "f_localtime");
#else
    vec->item[LT_GMTOFF].u.number = _timezone;
    vec->item[LT_ZONE].u.string = string_copy(_tzname[_daylight?1:0],"f_localtime");
#endif
#endif
#endif				/* sequent */
#endif				/* BSD42 */
    put_array(vec);
}
#endif

#ifdef F_RUSAGE
#ifdef WIN32
void f_rusage(void)
{
    error("rusage() not supported under Windows.\n");
}
#else

#ifdef RUSAGE
void
f_rusage(void)
{
    struct rusage rus;
    mapping_t *m;
    long usertime, stime;
    int maxrss;

    if (getrusage(RUSAGE_SELF, &rus) < 0) {
	m = allocate_mapping(0);
    } else {
#if 1 /* Was !SunOS_5 */
	usertime = rus.ru_utime.tv_sec * 1000 + rus.ru_utime.tv_usec / 1000;
	stime = rus.ru_stime.tv_sec * 1000 + rus.ru_stime.tv_usec / 1000;
#else
	usertime = rus.ru_utime.tv_sec * 1000 + rus.ru_utime.tv_nsec / 1000000;
	stime = rus.ru_stime.tv_sec * 1000 + rus.ru_stime.tv_nsec / 1000000;
#endif
	maxrss = rus.ru_maxrss;
#ifdef sun
	maxrss *= getpagesize() / 1024;
#endif
	m = allocate_mapping(16);
	add_mapping_pair(m, "utime", usertime);
	add_mapping_pair(m, "stime", stime);
	add_mapping_pair(m, "maxrss", maxrss);
	add_mapping_pair(m, "ixrss", rus.ru_ixrss);
	add_mapping_pair(m, "idrss", rus.ru_idrss);
	add_mapping_pair(m, "isrss", rus.ru_isrss);
	add_mapping_pair(m, "minflt", rus.ru_minflt);
	add_mapping_pair(m, "majflt", rus.ru_majflt);
	add_mapping_pair(m, "nswap", rus.ru_nswap);
	add_mapping_pair(m, "inblock", rus.ru_inblock);
	add_mapping_pair(m, "oublock", rus.ru_oublock);
	add_mapping_pair(m, "msgsnd", rus.ru_msgsnd);
	add_mapping_pair(m, "msgrcv", rus.ru_msgrcv);
	add_mapping_pair(m, "nsignals", rus.ru_nsignals);
	add_mapping_pair(m, "nvcsw", rus.ru_nvcsw);
	add_mapping_pair(m, "nivcsw", rus.ru_nivcsw);
    }
    push_refed_mapping(m);
}
#else

#ifdef GET_PROCESS_STATS
void
f_rusage(void)
{
    struct process_stats ps;
    mapping_t *m;
    int utime, stime, maxrss;

    if (get_process_stats(NULL, PS_SELF, &ps, NULL) == -1)
	m = allocate_mapping(0);
    else {
	utime = ps.ps_utime.tv_sec * 1000 + ps.ps_utime.tv_usec / 1000;
	stime = ps.ps_stime.tv_sec * 1000 + ps.ps_stime.tv_usec / 1000;
	maxrss = ps.ps_maxrss * getpagesize() / 1024;

	m = allocate_mapping(19);
	add_mapping_pair(m, "utime", utime);
	add_mapping_pair(m, "stime", stime);
	add_mapping_pair(m, "maxrss", maxrss);
	add_mapping_pair(m, "pagein", ps.ps_pagein);
	add_mapping_pair(m, "reclaim", ps.ps_reclaim);
	add_mapping_pair(m, "zerofill", ps.ps_zerofill);
	add_mapping_pair(m, "pffincr", ps.ps_pffincr);
	add_mapping_pair(m, "pffdecr", ps.ps_pffdecr);
	add_mapping_pair(m, "swap", ps.ps_swap);
	add_mapping_pair(m, "syscall", ps.ps_syscall);
	add_mapping_pair(m, "volcsw", ps.ps_volcsw);
	add_mapping_pair(m, "involcsw", ps.ps_involcsw);
	add_mapping_pair(m, "signal", ps.ps_signal);
	add_mapping_pair(m, "lread", ps.ps_lread);
	add_mapping_pair(m, "lwrite", ps.ps_lwrite);
	add_mapping_pair(m, "bread", ps.ps_bread);
	add_mapping_pair(m, "bwrite", ps.ps_bwrite);
	add_mapping_pair(m, "phread", ps.ps_phread);
	add_mapping_pair(m, "phwrite", ps.ps_phwrite);
    }
    push_refed_mapping(m);
}
#else

#ifdef TIMES			/* has times() but not getrusage() */

/*
  warning times are reported in processor dependent units of time.
  see man pages for 'times' to figure out how long a tick is on your system.
*/

void
f_rusage(void)
{
    mapping_t *m;
    struct tms t;

    times(&t);
    m = allocate_mapping(2);
    add_mapping_pair(m, "utime", t.tms_utime * 1000 / CLK_TCK);
    add_mapping_pair(m, "stime", t.tms_stime * 1000 / CLK_TCK);
    push_refed_mapping(m);
}

#else

#ifdef LATTICE

void
f_rusage(void)
{
    mapping_t *m;
    int i;
    unsigned int clock[2];

    i = timer(clock);		/* returns 0 if success, -1 otherwise */
    m = allocate_mapping(2);
    add_mapping_pair(m, "utime", i ? 0 : clock[0] * 1000 + clock[1] / 1000);
    add_mapping_pair(m, "stime", i ? 0 : clock[0] * 1000 + clock[1] / 1000);
    push_refed_mapping(m);
}

#endif				/* LATTICE */

#endif				/* TIMES */

#endif				/* GET_PROCESS_STATS */

#endif				/* RUSAGE */

#endif				/* WIN32 */

#endif

#ifdef F_MALLOC_CHECK
/* this efun only useful on the NeXT (func_spec.c has #ifdef NeXT).  A
   non-zero return value indicates that some memory corruption has occurred
   at some time prior to this calling of this efun.
*/
void
f_malloc_check(void)
{
    push_number(NXMallocCheck());
}
#endif

#ifdef F_MALLOC_DEBUG
/* NeXT specific efun for setting the debugging level of NeXT's built-in
   malloc.
*/
void
f_malloc_debug(void)
{
    int level;

    level = sp->u.number;
    if (level < 0) {
        int rc;

        rc = malloc_debug(0);
        malloc_singlethreaded();
        sp->u.number = rc;
    } else {
        sp->u.number = malloc_debug(level);
    }
}
#endif
