/*  $Id$

    This program is a part of Neolith project distribution. The Neolith
    project is based on MudOS v22pre5 LPmud driver. Read doc/Copyright
    before you try to use, modify or distribute this program.

    For more information about Neolith project, please visit:

    http://www.es2.muds.neet/neolith
 */

#ifndef OBJECT_H
#define OBJECT_H

/*
 * Definition of an object.
 * If the object is inherited, then it must not be destructed !
 *
 * The reset is used as follows:
 * 0: There is an error in the reset() in this object. Never call it again.
 * 1: Normal state.
 * 2 or higher: This is an interactive user, that has not given any commands
 *		for a number of reset periods.
 */

#include "uids.h"

#define MAX_OBJECT_NAME_SIZE	2048

#define O_HEART_BEAT		0x0001	/* Does it have an heart beat ?      */
#define O_IS_WIZARD		0x0002	/* used to be O_IS_WIZARD            */
#define O_LISTENER              0x0004	/* can hear say(), etc */
#define O_ENABLE_COMMANDS	0x0004	/* Can it execute commands ?         */
#define O_CLONE			0x0008	/* Is it cloned from a master copy ? */
#define O_DESTRUCTED		0x0010	/* Is it destructed ?                */
#define O_SWAPPED		0x0020	/* Is it swapped to file             */
#define O_ONCE_INTERACTIVE	0x0040	/* Has it ever been interactive ?    */
#define O_RESET_STATE		0x0080	/* Object in a 'reset':ed state ?    */
#define O_WILL_CLEAN_UP		0x0100	/* clean_up will be called next time */
#define O_VIRTUAL		0x0200	/* We're a virtual object            */
#define O_HIDDEN		0x0400	/* We're hidden from nonprived objs  */
#define O_EFUN_SOCKET           0x0800	/* efun socket references object     */
#define O_WILL_RESET            0x1000	/* reset will be called next time    */
#define O_UNUSED                0x8000

/*
 * Note: use of more than 16 bits means extending flags to an unsigned long
 */

typedef struct sentence_s {
    char *verb;
    struct sentence_s *next;
    struct object_s *ob;
    union string_or_func function;
    int flags;
} sentence_t;

typedef struct object_s {
    unsigned short ref;		/* Reference count. */
    unsigned short flags;	/* Bits or'ed together from above */
    char *name;
    struct object_s *next_hash;
    /* the fields above must match lpc_object_t */
    int load_time;		/* time when this object was created */
    int next_reset;		/* Time of next reset of this object */
    int time_of_ref;		/* Time when last referenced. Used by swap */
    long swap_num;		/* Swap file offset. -1 is not swapped yet. */
    program_t *prog;
    struct object_s *next_all;
    struct object_s *next_inv;
    struct object_s *contains;
    struct object_s *super;	/* Which object surround us ? */
    struct interactive_s *interactive;	/* Data about an interactive user */
    sentence_t *sent;
    struct object_s *next_hashed_living;
    char *living_name;		/* Name of living object if in hash */
    userid_t *uid;		/* the "owner" of this object */
    userid_t *euid;		/* the effective "owner" */
    svalue_t variables[1];	/* All variables to this program */
    /* The variables MUST come last in the struct */
} object_t;

#define add_ref(ob, str) ob->ref++

#define ROB_STRING_ERROR 1
#define ROB_ARRAY_ERROR 2
#define ROB_MAPPING_ERROR 4
#define ROB_NUMERAL_ERROR 8
#define ROB_GENERAL_ERROR 16
#define ROB_CLASS_ERROR 32
#define ROB_ERROR 63

extern object_t *hashed_living[CFG_LIVING_HASH_SIZE];
extern object_t *previous_ob;
extern int tot_alloc_object;
extern int tot_alloc_object_size;
extern int save_svalue_depth;

void bufcat(char **, char *);
inline int svalue_save_size(svalue_t *);
inline void save_svalue(svalue_t *, char **);
inline int restore_svalue(char *, svalue_t *);
int save_object(object_t *, char *, int);
char *save_variable(svalue_t *);
int restore_object(object_t *, char *, int);
void restore_variable(svalue_t *, char *);
object_t *get_empty_object(int);
void reset_object(object_t *);
void call_create(object_t *, int);
void reload_object(object_t *);
void free_object(object_t *, char *);
object_t *find_living_object(char *, int);
inline int valid_hide(object_t *);
inline int object_visible(object_t *);
void set_living_name(object_t *, char *);
void remove_living_name(object_t *);
void stat_living_objects(outbuffer_t *);
void tell_npc(object_t *, char *);
void tell_object(object_t *, char *);
int find_global_variable(program_t *, char *, unsigned short *);
void dealloc_object(object_t *, char *);

#endif	/* OBJECT_H */
