/*  $Id$

    This program is a part of Neolith project distribution. The Neolith
    project is based on MudOS v22pre5 LPmud driver. Read doc/Copyright
    before you try to use, modify or distribute this program.

    For more information about Neolith project, please visit:

    http://www.es2.muds.neet/neolith
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif	/* HAVE_CONFIG_H */

#ifdef	STDC_HEADERS
#include <locale.h>
#endif	/* STDC_HEADERS */

#define SUPPRESS_COMPILER_INLINES
#include "std.h"
#include "lpc_incl.h"
#include "backend.h"
#include "simul_efun.h"
#include "binaries.h"
#include "main.h"
#include "otable.h"
#include "comm.h"
#include "main.h"

port_def_t external_port[5];

static int e_flag = 0;		/* Load empty, without preloads. */
int t_flag = 0;			/* Disable heart beat and reset */
int comp_flag = 0;		/* Trace compilations */
int max_cost;
int time_to_swap;
int time_to_clean_up;
char *default_fail_message;
int boot_time;
int max_array_size;
int max_buffer_size;
int max_string_length;
char *master_file_name;
static int reserved_size;
char *reserved_area;		/* reserved for MALLOC() */
static char *mud_lib;

svalue_t const0, const1, const0u;

double consts[NUM_CONSTS];

/* -1 indicates that we have never had a master object.  This is so the
 * simul_efun object can load before the master. */
object_t *master_ob = 0;

static RETSIGTYPE sig_fpe(int sig);
static RETSIGTYPE sig_cld(int sig);

static RETSIGTYPE sig_usr1(int sig);
static RETSIGTYPE sig_usr2(int sig);
static RETSIGTYPE sig_term(int sig);
static RETSIGTYPE sig_int(int sig);

static RETSIGTYPE sig_hup(int sig);
static RETSIGTYPE sig_abrt(int sig);
static RETSIGTYPE sig_segv(int sig);
static RETSIGTYPE sig_ill(int sig);
static RETSIGTYPE sig_bus(int sig);
static RETSIGTYPE sig_iot(int sig);

int
main(int argc, char ** argv)
{
    time_t tm;
    int i, new_mudlib = 0, got_defaults = 0;
    int no_ip_demon = 0;
    char *p;
    char version_buf[80];
    error_context_t econ;
    struct lpc_predef_s predefs;

#if defined(ENABLE_NLS) && !defined(__CYGWIN__)
    setlocale(LC_ALL, "");
    bindtextdomain(PACKAGE, LOCALEDIR);
    textdomain(PACKAGE);
#endif	/* ENABLE_NLS */

    tzset();	/* initialize time zone information */
    boot_time = time(NULL);
    srand(boot_time);	/* lƶüƲ; */

    const0.type = T_NUMBER;
    const0.u.number = 0;
    const1.type = T_NUMBER;
    const1.u.number = 1;

    /* const0u used by undefinedp() */
    const0u.type = T_NUMBER;
    const0u.subtype = T_UNDEFINED;
    const0u.u.number = 0;

    fake_prog.program_size = 0;

    /* * Check that the definition of EXTRACT_UCHAR() is correct. */
    p = (char *) &i;
    *p = -10;
    if (EXTRACT_UCHAR(p) != 0x100 - 10) {
	fprintf(stderr, "Bad definition of EXTRACT_UCHAR() in interpret.h.\n");
	exit(-1);
    }

    /*
     * An added test: can we do EXTRACT_UCHAR(x++)?
     * (read_number, etc uses it)
     */
    p = (char *) &i;
    (void) EXTRACT_UCHAR(p++);
    if ((p - (char *) &i) != 1) {
	fprintf(stderr, "EXTRACT_UCHAR() in interpret.h evaluates its argument more than once.\n");
	exit(-1);
    }

    /* Check the living hash table size */
    if (CFG_LIVING_HASH_SIZE != 4 && CFG_LIVING_HASH_SIZE != 16 &&
	CFG_LIVING_HASH_SIZE != 64 && CFG_LIVING_HASH_SIZE != 256 &&
	CFG_LIVING_HASH_SIZE != 1024 && CFG_LIVING_HASH_SIZE != 4096) {
	fprintf(stderr, "CFG_LIVING_HASH_SIZE in options.h must be one of 4, 16, 64, 256, 1024, 4096, ...\n");
	exit(-1);
    }

    current_time = time(NULL);

    /* read in the configuration file */
    got_defaults = 0;
    for (i = 1; (i < argc) && !got_defaults; i++) {
	if (argv[i][0] != '-') {
	    set_defaults(argv[i]);
	    got_defaults = 1;
	}
    }

    get_version(version_buf);
    if (!got_defaults) {
	fprintf(stderr, "%s\n", version_buf);
	fprintf(stderr, _("usage: %s config-file\n"), argv[0]);
	exit(-1);
    }

    printf(_("Initializing ....\n"));
    init_strings();		/* in stralloc.c */
    init_otable();		/* in otable.c */
    init_identifiers();		/* in lex.c */
    init_locals();              /* in compiler.c */

    time_to_clean_up = TIME_TO_CLEAN_UP;
    time_to_swap = TIME_TO_SWAP;
    max_cost = MAX_COST;
    reserved_size = RESERVED_SIZE;
    max_array_size = MAX_ARRAY_SIZE;
    max_buffer_size = MAX_BUFFER_SIZE;
    max_string_length = MAX_STRING_LENGTH;
    master_file_name = (char *) MASTER_FILE;

    /* fix the filename */
    while (*master_file_name == '/') master_file_name++;
    p = master_file_name;
    while (*p++);
    if (p[-2]=='c' && p[-3]=='.')
	p[-3]=0;

    mud_lib = (char *) MUD_LIB;
    set_inc_list(INCLUDE_DIRS);
    if (reserved_size > 0)
	reserved_area = (char *) DMALLOC(reserved_size, TAG_RESERVED, "main.c: reserved_area");
    for (i = 0; i < sizeof consts / sizeof consts[0]; i++)
	consts[i] = exp(-i / 900.0);
    init_num_args();
    reset_machine(1);
    /*
     * The flags are parsed twice ! The first time, we only search for the -m
     * flag, which specifies another mudlib, and the D-flags, so that they
     * will be available when compiling master.c.
     */
    for (i = 1; i < argc; i++) {
	if (argv[i][0] != '-')
	    continue;
	switch (argv[i][1]) {
	case 'D':
	    if (argv[i][2]) {	/* Amylaar : allow flags to be passed down to
				 * the LPC preprocessor */
		struct lpc_predef_s *tmp;

		tmp = &predefs;
		tmp->flag = argv[i] + 2;
		tmp->next = lpc_predefs;
		lpc_predefs = tmp;
		continue;
	    }
	    fprintf(stderr, "Illegal flag syntax: %s\n", argv[i]);
	    exit(-1);
	case 'N':
	    no_ip_demon++;
	    continue;
#ifdef YYDEBUG
	case 'y':
	    yydebug = 1;
	    continue;
#endif				/* YYDEBUG */
	case 'm':
	    mud_lib = alloc_cstring(argv[i] + 2, "mudlib dir");
	    if (chdir(mud_lib) == -1) {
		fprintf(stderr, "Bad mudlib directory: %s\n", mud_lib);
		exit(-1);
	    }
	    new_mudlib = 1;
	    break;
	}
    }
    if (!new_mudlib && chdir(mud_lib) == -1) {
	fprintf(stderr, "Bad mudlib directory: %s\n", mud_lib);
	exit(-1);
    }
    time(&tm);
    debug_message(_("%s (%s) starting up - %s\n\n"),
	MUD_NAME, version_buf, ctime(&tm));

    init_binaries(argc, argv);
    add_predefines();

    eval_cost = max_cost;	/* needed for create() functions */

    save_context(&econ);
    if (setjmp(econ.context)) {
	debug_message("The simul_efun (%s) and master (%s) objects must be loadable.\n",
	    SIMUL_EFUN, MASTER_FILE);
	exit(-1);
    } else {
	init_simul_efun(SIMUL_EFUN);
	init_master(MASTER_FILE);
    }
    pop_context(&econ);

    for (i = 1; i < argc; i++) {
	if (argv[i][0] != '-') {
	    continue;
	} else {
	    /*
	     * Look at flags. -m and -o has already been tested.
	     */
	    switch (argv[i][1]) {
	    case 'D':
	    case 'N':
	    case 'm':
	    case 'y':
		continue;
	    case 'f':
		save_context(&econ);
		if (setjmp(econ.context)) {
		    debug_message("Error while calling master::flag(\"%s\"), aborting ...\n", argv[i] + 2);
		    exit(-1);
		}
		push_constant_string(argv[i] + 2);
		(void) apply_master_ob(APPLY_FLAG, 1);
		if (MudOS_is_being_shut_down) {
		    debug_message("Shutdown by master object.\n");
		    exit(0);
		}
		pop_context(&econ);
		continue;
	    case 'e':
		e_flag++;
		continue;
	    case 'p':
		external_port[0].port = atoi(argv[i] + 2);
		continue;
	    case 'c':
		comp_flag++;
		continue;
	    case 't':
		t_flag++;
		continue;
	    default:
		debug_message("Unknown flag: %s\n", argv[i]);
		exit(-1);
	    }
	}
    }

    if (MudOS_is_being_shut_down) exit(1);

    if (*(DEFAULT_FAIL_MESSAGE)) {
	char buf[8192];

	strcpy(buf, DEFAULT_FAIL_MESSAGE);
	strcat(buf, "\n");
	default_fail_message = make_shared_string(buf);
    } else
	default_fail_message = "What?\n";

    preload_objects(e_flag);

#ifdef SIGFPE
    signal(SIGFPE, sig_fpe);
#endif
#ifdef SIGUSR1
    signal(SIGUSR1, sig_usr1);
#endif
#ifdef SIGUSR2
    signal(SIGUSR2, sig_usr2);
#endif
    signal(SIGTERM, sig_term);
    signal(SIGINT, sig_int);
#if defined(SIGABRT) && !defined(LATTICE)
    signal(SIGABRT, sig_abrt);
#endif
#ifdef SIGIOT
    signal(SIGIOT, sig_iot);
#endif
#ifdef SIGHUP
    signal(SIGHUP, sig_hup);
#endif
#ifdef SIGBUS
    signal(SIGBUS, sig_bus);
#endif
#ifndef LATTICE
    signal(SIGSEGV, sig_segv);
    signal(SIGILL, sig_ill);
#endif
#ifdef USE_BSD_SIGNALS
    signal(SIGCHLD, sig_cld);
#else
    signal(SIGCLD, sig_cld);
#endif

    backend();

    return 0;
}

void
debug_message(char* fmt, ...)
{
    static int append = 0;
    static char deb_buf[100];
    static char *deb = deb_buf;
    va_list args;
    FILE *fp = NULL;

    if (!append) {
	/*
	 * check whether config file specified this option
	 */
	if (strlen(DEBUG_LOG_FILE))
	    sprintf(deb, "%s/%s", LOG_DIR, DEBUG_LOG_FILE);
	else
	    sprintf(deb, "%s/debug.log", LOG_DIR);
	while (*deb == '/')
	    deb++;
    }
    fp = fopen(deb, append ? "a" : "w");

    /*
     * re-use stdout's file descriptor if system or process runs out
     */
    if ( !fp && (errno == EMFILE || errno == ENFILE)) {
	fp = freopen(deb, append ? "a" : "w", stdout);
	append = 2;
    }
    if ( !fp ) {
	/* darn.  We're in trouble */
	perror(deb);
	abort();
    }

    va_start(args, fmt);
    vfprintf(fp, fmt, args);
    fflush(fp);
    vfprintf(stderr, fmt, args);
    fflush(stderr);
    va_end(args);

    /* don't close stdout */
    if (append != 2)
	(void) fclose(fp);

    /* append to debug.log next time thru */
    if (!append) append = 1;
}

void
debug_perror(char* what, char* file)
{
    if (file)
	debug_message("System Error: %s:%s:%s\n", what, file, strerror(errno));
    else
	debug_message("System Error: %s:%s\n", what, strerror(errno));
}

int slow_shut_down_to_do = 0;

char *
xalloc(int size)
{
    char *p;
    static int going_to_exit;

    if (going_to_exit)
	exit(3);
    p = (char *) DMALLOC(size, TAG_MISC, "main.c: xalloc");
    if (p == 0) {
	if (reserved_area) {
	    FREE(reserved_area);
	    p = "Temporarily out of MEMORY. Freeing reserve.\n";
	    write(1, p, strlen(p));
	    reserved_area = 0;
	    slow_shut_down_to_do = 6;
	    return xalloc(size);/* Try again */
	}
	going_to_exit = 1;
	fatal("Totally out of MEMORY.\n");
    }
    return p;
}

static RETSIGTYPE
sig_cld(int sig)
{
    int status;

    while (wait3(&status, WNOHANG, NULL) > 0)
	;
}


static RETSIGTYPE
sig_fpe(int sig)
{
    signal(SIGFPE, sig_fpe);
}

/* send this signal when the machine is about to reboot.  The script
   which restarts the MUD should take an exit code of 1 to mean don't
   restart
 */

static RETSIGTYPE
sig_usr1(int sig)
{
    push_constant_string("Host machine shutting down");
    push_undefined();
    push_undefined();
    apply_master_ob(APPLY_CRASH, 3);
    debug_message("Received SIGUSR1, calling exit(-1)\n");
    exit(-1);
}

/* Abort evaluation */
static RETSIGTYPE
sig_usr2(int sig)
{
    eval_cost = 1;
}

/*
 * Actually, doing all this stuff from a signal is probably illegal
 * -Beek
 */
static RETSIGTYPE
sig_term(int sig)
{
    fatal("Process terminated");
}

static RETSIGTYPE
sig_int(int sig)
{
    fatal("Process interrupted");
}

static RETSIGTYPE
sig_segv(int sig)
{
    fatal("Segmentation fault");
}

static RETSIGTYPE
sig_bus(int sig)
{
    fatal("Bus error");
}

static RETSIGTYPE
sig_ill(int sig)
{
    fatal("Illegal instruction");
}

static RETSIGTYPE
sig_hup(int sig)
{
    fatal("Hangup!");
}

static RETSIGTYPE
sig_abrt(int sig)
{
    fatal("Aborted");
}

static RETSIGTYPE
sig_iot(int sig)
{
    fatal("Aborted(IOT)");
}

